""" Provides distribution functionality based on a named dictionary. """

#-------------------------------------------------------------------------------
import numpy as np
import collections
from probayes.vtypes import issingleton
from probayes.named_dict import NamedDict

#-------------------------------------------------------------------------------
class Distribution (NamedDict):
  """ A distribution is a named topological space defined over specific values 
  for one or more variables. Those values are defined by the dictionary
  {variable_name: values} [Distribution.values()] with dimensions 
  {variable_name: dim} [Distribution.dims]. The dimensionality defines a finite 
  space of sizes with it permissable for variables to share dimensions and in 
  such cases reducing the shape dimensions.

  :example:
  >>> import numpy as np
  >>> import probayes as pb
  >>> xy_dict = {'x': np.arange(3), 'y': np.arange(2)}
  >>> xy = pb.Distribution('x,y', xy_dict, dims={'x': 0, 'y': 1})
  >>> print(xy.shape)
  [3, 2]

  Scalars can be included among vals, but must possess dimensionality None and
  do not contribute to the size or shape of the distribution.
  """

  # Protected
  _dims = None         # Ordered dictionary specifying dimension index of vals
  _ndim = None         # Number of dimensions
  _sizes = None        # Size of dimensions including shared
  _shape = None        # Size of dimension shape excluding shared
  _size = None         # prod(sizes)
  _keylist = None      # Keys of vals as list
  _keyset = None       # Keys of vals a set
  _aresingleton = None # Whether vals are scalars
  _issingleton = None  # all(_aresingleton)
  _disallowed = {'attrs', 'prob', 'pscale'} # disallowed keys

#-------------------------------------------------------------------------------
  def __init__(self, name, *args, **kwds):
    """ Initialises the Distribution with a name, args, and kwds in the same
    way as NamedDict(), except with 'dims' as a reserved keyword to set the
    dimensionality. """
    args = tuple(args)
    kwds = dict(kwds)
    dims = None if 'dims' not in kwds else kwds.pop('dims')
    super().__init__(name, *args, **kwds)
    self.dims = dims
  
#-------------------------------------------------------------------------------
  @property
  def dims(self):
    return self._dims

  @property
  def ndim(self):
    return self._ndim

  @property
  def sizes(self):
    return self._sizes

  @property
  def shape(self):
    return self._shape

  @property
  def size(self):
    return self._size

  @property
  def keylist(self):
    return self._keylist

  @property
  def short_name(self):
    return ','.join(self._keylist)

  @property
  def keyset(self):
    return self._keyset

  @property
  def aresingleton(self):
    return self._aresingleton

  @property
  def issingleton(self):
    return self._issingleton

  @dims.setter
  def dims(self, dims=None):
    """ Sets the dimensions for each of the variables.

    :param dims: a dictionary of {variable_name: variable_dim}

    The keys should correspond to that of a dictionary. If dims
    is None, then the dimensionality is set according to the
    order in values.

    :example:
    >>> import numpy as np
    >>> from collections import OrderedDict()
    >>> import probayes as pb
    >>> xy_dict = {'x': np.arange(3), 'y': np.arange(2)}
    >>> xy = pb.Distribution('x,y', xy_dict)
    >>> print(xy.dims))
    OrderedDict([('x', 0), ('y', 1)])
    """
    self._dims = dims
    self._sizes = []
    self._shape = []
    self._size = None
    self._keylist = list(self.keys())
    self._keyset = frozenset(self._keylist)
    self._aresingleton = []
    self._issingleton = None
    eval_dims = self._dims is None
    if eval_dims:
      self._dims = collections.OrderedDict()
    else:
      self._dims = collections.OrderedDict(self._dims)
    if not self.__len__():
      return

    # Tranform {None} to {0} to play nicely with isunitsetint
    for key in self._keylist:
      if key in self._disallowed:
        raise ValueError(f"Disallowed key: {key}")
      if isinstance(self[key], set):
        if len(self[key]) == 1:
          element = list(self[key])[0]
          if element is None:
            self.update({key: {0}})

    # Count number of non-singleton dimensions
    self._aresingleton = [issingleton(val) for val in self.values()]
    self._issingleton = np.all(self._aresingleton)
    self._ndim = 0 
    for dim in self._dims.values():
      if dim is not None:
        self._ndim = max(self._ndim, dim+1)

    # Corroborate vals and dims
    ones_ndim = np.ones(self._ndim, dtype=int)
    self._shape = [None] * self.ndim
    nonsingleton_count = -1
    for i, key in enumerate(self._keylist):
      values = self[key]

      # Scalars are dimensionless and therefore shapeless
      if self._aresingleton[i]:
        if eval_dims:
          self._dims.update({key:None})
        elif key in self._dims:
          assert self._dims[key] == None,\
            "Dimension index for scalar value {} must be None, not {}".\
            format(key, self._dims[key])
        else:
          self._dims.update({key: None})

      # Non-scalars require correct dimensionality
      else:
        nonsingleton_count += 1
        assert isinstance(values, np.ndarray), \
            "Dictionary of numpy arrays expected for nonsingletons but found" + \
            "type {} for key {}".format(type(values), key)
        val_size = values.size
        assert val_size == np.max(values.shape), \
            "Values must have one non-singleton dimension but found" + \
            "shape {} for key {}".format(values.shape, key)
        if eval_dims:
          self._dims.update({key: nonsingleton_count})
          if len(self._shape) == nonsingleton_count:
            self._shape.append(None)
            self._ndim = len(self._shape)
            ones_ndim = np.ones(self._ndim, dtype=int)
        else:
          assert key in self._dims, "Missing key {} in dims specification {}".\
              format(key, self._dims)
        self._sizes.append(val_size)
        self._shape[self._dims[key]] = val_size
        vals_shape = np.copy(ones_ndim)
        vals_shape[self._dims[key]] = val_size
        re_shape = self._ndim != values.ndim or \
                   any(np.array(values.shape) != vals_shape)
        if re_shape:
          self[key] = values.reshape(vals_shape)
    self._size = int(np.prod(self._shape))

#-------------------------------------------------------------------------------
  def singleton(self, key=None):
    """ Returns whether Distribution defines a singleton according key:

    if key is None (default): returns self.issingleton
    if key is an int: returns self.aresingleton[key]
    if key is an key: returns key-corresponded element of self.aresingleton
    """
    if key is None:
      return self._issingleton
    if isinstance(key, str):
      if key not in self._keys:
        return None
      key = self._keys.index(key)
    return self._issingleton[key]

#-------------------------------------------------------------------------------
  def lookup(self, keys, as_dist=False):
    """ Returns the values of Distribution filtering by keys """ 
    keys = keys or self._keyset
    keys = set(keys)
    vals = collections.OrderedDict()
    if as_dist:
      dims = collections.OrderedDict()
      keylist = []
      for key in self._keylist:
        if key in keys:
          keylist.append(key)
          vals.update({key: self[key]})
          dims.update({key: self._dims[key]})
      return Distribution(','.join(keylist), vals, dims=dims)
    seen_keys = set()
    for i, key in enumerate(self._keylist):
      if key in keys and key not in seen_keys:
        if self._aresingleton[i]:
          seen_keys.add(key)
          vals.update({key: self[key]})
        else:
          shared_keys = [key]
          for j, cand_key in enumerate(self._keylist):
            if j > i and cand_key in keys and not self._aresingleton[j]:
              if self.dims[key] == self.dims[cand_key]:
                shared_keys.append(cand_key)
          if len(shared_keys) == 1:
            vals.update({key: np.ravel(self[key])})
            seen_keys.add(key)
          else:
            val = [None] * len(shared_keys)
            for j, shared_key in enumerate(shared_keys):
              val[j] = np.ravel(self[shared_key])
              seen_keys.add(shared_key)
            vals.update({','.join(shared_keys): tuple(val)})
    return vals

#-------------------------------------------------------------------------------
  def redim(self, dims):
    """  Returns a Distribution according to redimensionised values in dims, 
    index-ordered by the order in dims.
    """
    for key in self._keylist:
      if self._dims[key] is not None:
        assert key in dims, \
            "Missing key for nonsingleton {} in dim {}".format(key, dims)
      elif key in dims:
        assert dims[key] is None, \
            "Dimension {} requested for singleton with key {}".\
            format(dims[key], key)
    vals = {key: self[key] for key in dims.keys()}
    return Distribution(self._name, vals, dims=dims)

#-------------------------------------------------------------------------------
  def rekey(self, keymap):
    """ Returns a Manifold rekeying the values and dimensions according the
    dictionary mappings given in keymap. 
    """
    assert isinstance(keymap, dict), \
        "Input keymap must a dictionary in the form {old_key: new_key}"
    vals = collections.OrderedDict()
    dims = collections.OrderedDict()
    for key in self._keylist:
      if key not in keymap.keys():
        vals.update({key: self[key]})
        dims.update({key: self.dims[key]})
      else:
        map_key = keymap[key]
        assert map_key not in vals, \
            "Key map results in duplicate for key: {}".format(map_key)
        vals.update({map_key: self[key]})
        dims.update({map_key: self.dims[key]})
    return Distribution(self._name, vals, dims=dims)
    
#-------------------------------------------------------------------------------
  def serialise(self):
    short_name = self.short_name
    serialised = {short_name: {key:val for key, val in self.items()}}
    serialised[short_name].update({'attrs': self.dims})
    return serialised

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



import numpy as np
from probayes.vtypes import isunitset, isscalar, uniform, eval_vtype, VTYPES
from probayes.pscales import eval_pscale, iscomplex, NEARLY_NEGATIVE_INF
"""
A module to provide functional support to rv.py
"""
#-------------------------------------------------------------------------------
def uniform_prob(*args, prob=None, inside=None, pscale=1.):
  """ Uniform probability function for discrete and continuous vtypes. """

  # Detect ptype, default to prob if no values, otherwise detect vtype  
  assert len(args) >= 1, "Minimum of a single positional argument"
  pscale = eval_pscale(pscale)
  use_logs = iscomplex(pscale)
  if prob is None:
    prob = 0. if use_logs else 1.
  vals = args[0]
  if vals is None:
    return prob
  vtype = eval_vtype(vals) if callable(inside) else eval_vtype(inside)

  # Set inside function by vtype if not specified
  if not callable(inside):
    if vtype in VTYPES[float]:
      inside = lambda x: np.logical_and(x >= min(inside), x <= max(inside))
    else:
      inside = lambda x: np.isin(x, inside)

  # If scalar, check within variable set
  p_zero = NEARLY_NEGATIVE_INF if use_logs else 0.
  if isscalar(vals):
    prob = prob if inside(vals) else p_zero

  # Otherwise treat as uniform within range
  else:
    p_true = prob
    prob = np.tile(p_zero, vals.shape)
    prob[inside(vals)] = p_true

  # This section below is there just to play nicely with conditionals
  if len(args) > 1:
    for arg in args[1:]:
      if use_logs:
        prob = prob + uniform_prob(arg, inside=inside, pscale=0.j)
      else:
        prob = prob * uniform_prob(arg, inside=inside)
  return prob

#-------------------------------------------------------------------------------
def matrix_cond_sample(pred_vals, succ_vals, prob, vset=None):
  """ Returns succ_vals with sampling """
  if not isunitset(succ_vals):
    return succ_vals
  assert isscalar(pred_vals), \
      "Can only cumulatively sample from a single predecessor"
  assert prob.ndim==2 and len(set(prob.shape)) == 1, \
      "Transition matrix must be a square"
  support = prob.shape[0]
  if vset is None:
    vset = set(range(support))
  else:
    assert len(vset) == support, \
        "Transition matrix size {} incommensurate with set support {}".\
        format(support, len(vset))
  vset = sorted(vset)
  pred_idx = vset.index(pred_vals)
  cmf = np.cumsum(prob[:, pred_idx], axis=0)
  succ_cmf = list(succ_vals)[0]
  if type(succ_cmf) in VTYPES[int]:
    succ_cmf = uniform(0., 1., succ_cmf)
  else:
    succ_cmf = np.atleast_1d(succ_cmf)
  succ_idx = np.maximum(0, np.minimum(support-1, np.digitize(succ_cmf, cmf)))
  return vset[succ_idx], pred_idx, succ_idx

#-------------------------------------------------------------------------------
def lookup_square_matrix(col_vals, row_vals, sq_matrix, 
                         vset=None, col_idx=None, row_idx=None):
  assert sq_matrix.ndim==2 and len(set(sq_matrix.shape)) == 1, \
      "Transition matrix must be a square"
  support = sq_matrix.shape[0]
  if vset is None:
    vset = list(range(support))
  else:
    assert len(vset) == support, \
        "Transition matrix size {} incommensurate with set support {}".\
        format(support, len(vset))
    vset = sorted(vset)
  rc_scalar = False
  if row_idx is None:
    if isscalar(row_vals):
      rc_scalar = True
      row_idx = vset.index(row_vals)
    else:
      if isinstance(row_vals, np.ndarray):
        row_vals = np.ravel(row_vals).tolist()
      row_idx = [vset.index(row_val) for row_val in row_vals]
  elif isscalar(row_idx):
      rc_scalar = True
  if col_idx is None:
    if isscalar(col_vals):
      rc_scalar = True
      col_idx = vset.index(col_vals)
    else:
      if isinstance(col_vals, np.ndarray):
        col_vals = np.ravel(col_vals).tolist()
      col_idx = [vset.index(col_val) for col_val in col_vals]
  elif isscalar(col_idx):
      rc_scalar = True
  if rc_scalar:
    return sq_matrix[row_idx, col_idx]
  mat = np.empty([len(row_idx), len(col_idx)], dtype=sq_matrix.dtype)
  for i, row in enumerate(row_idx):
    mat[i] = sq_matrix[row][col_idx]
  return mat
    
#-------------------------------------------------------------------------------



""" Provides Random Field class: RF """

#-------------------------------------------------------------------------------
import collections
import numpy as np

from probayes.field import Field
from probayes.rv import RV
from probayes.prob import Prob
from probayes.pd import PD
from probayes.pd_utils import margcond_str
from probayes.vtypes import isscalar, isunitsetint
from probayes.pscales import iscomplex, prod_pscale
from probayes.rf_utils import rv_prod_rule, sample_cond_cov
from probayes.expr import Expr
from probayes.expression import Expression
from probayes.cf import CF
from probayes.cond_cov import CondCov

DEFAULT_CONDITIONAL_PROBABILITY = {False: 1., True: 0.}

#-------------------------------------------------------------------------------
class RF (Field, Prob):
  """
  A random field is a collection of a variables, that includes at least one RV, 
  that participate in a joint probability distribution function without 
  explicit directional conditionality.
  
  Since this class is intended as a building block for SD instances and networkx 
  cannot mix undirected and directed graphs, edges cannot be defined explicitly 
  within this class. Use SD if directed edges are required. Implicit support for
  undirected edges is provided by the set_prob(), set_prop(), and set_tran()
  methods.
  """

  # Protected
  _defiid = None     # Default IID random variables for calling distributions
  _passdims = None   # Flag to pass dimensions to probability function
  _prop = None       # Non-transitional proposition function
  _prop_deps = None  # Set of proposition dependencies
  _tran = None       # Transitional proposition function
  _tfun = None       # CDF/IDF of transition function 
  _tsteps = None     # Number of steps per transitional modificiation
  _cvars = None      # Conditional random variable sampling specification
  _sym_tran = None   # Flag for symmetrical transitional conditional functions
  _is_stochastic = True # Flag of whether stochastic

  # Private
  __def_prob = None   # Flag to denote prob is defaulted
  __cond_mod = None  # conditional RV index modulus
  __cond_cov = None  # conditional covariance matrix

#-------------------------------------------------------------------------------
  def __init__(self, *args): # over-rides NX_GRAPH.__init__()
    """ Initialises a random field with RVs for in args. See set_vars(). """
    super().__init__(*args)
    self.set_prob()
    self.set_prop()
    self.set_tran()

#-------------------------------------------------------------------------------
  def add_cd(self, *args, **kwds):
    """ Adding of conditional dependences disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  def _refresh(self):
    """ Updates RV summary objects, RF name and id, and delta factory. """
    super()._refresh()
    if self._nvars:
      pscales = [var.pscale for var in self._vars.values()]
      self.pscale = prod_pscale(pscales)
    self.eval_length()
    self._prop = None
    self._tran = None
    self._tfun = None

#-------------------------------------------------------------------------------
  @property
  def prob(self):
    return self._prob

  @property
  def passdims(self):
    return self.passdims

  @property
  def def_prob(self):
    return self.__def_prob

  def set_prob(self, prob=None, *args, **kwds):
    """ Sets the joint probability with optional arguments and keywords.

    :param prob: may be a scalar, array, or callable function.
    :param *args: optional arguments to pass if prob is callable.
    :param **kwds: optional keywords to pass if prob is callable.

    'pscale' is a reserved keyword. See Prob.pscale for explanation of how 
    pscale is used. Keyword 'passdims' is reserved to pass a flag to callable
    prob functions to pass the dimensionality dictionary for values.
    """
    kwds = dict(kwds)
    self._passdims = False if 'passdims' not in kwds else kwds.pop('passdims')
    if 'pscale' not in kwds and self._nvars:
      pscales = [var.pscale for var in self._varlist]
      kwds.update({'pscale': prod_pscale(pscales)})
    super().set_prob(prob, *args, **kwds)
    if self._prob is None:
      self._default_prob()
      return
    assert self._anystoch, \
        "Cannot set field probability without any random variables"
    self.__def_prob = False

#-------------------------------------------------------------------------------
  def _default_prob(self):
    if not self._anystoch or (self._prob is not None and not self.__def_prob):
      return
    if all([var.isiconic for var in self._varlist]):
      prob = None
      for var in self._varlist:
        if var.is_stochastic:
          var_prob = var.prob
          mul = var_prob if not isinstance(var_prob, Expr) else var_prob._expr
          if prob is not None:
            prob = prob * mul
          else:
            prob = mul
      self.set_prob(prob, pscale=self._pscale)
      self.__def_prob = True

#-------------------------------------------------------------------------------
  def eval_length(self):
    """ Evaluates and returns the joint length of the field. """
    self._lengths = np.array([var.length for var in self._varlist \
                              if var.is_stochastic], dtype=float)
    self._length = np.sqrt(np.sum(self._lengths**2))
    return self._length

#-------------------------------------------------------------------------------
  @property
  def prop(self):
    """ Returns the proposal probability expression if specified """
    return self._prop

  def set_prop(self, prop=None, *args, **kwds):
    """ Sets the joint proposition function with optional arguments and keywords.

    :param prop: may be a scalar, array, or callable function.
    :param *args: optional arguments to pass if prop is callable.
    :param **kwds: optional keywords to pass if prop is callable.
    """
    self._prop = prop
    self._prop_deps = self._keylist if 'deps' not in kwds else kwds.pop['deps']
    if self._prop is None:
      return
    assert self._anystoch, \
        "Cannot set field proposition without any random variables"
    assert self._tran is None, \
        "Cannot assign both proposition and transition probabilities"
    self._prop = Expression(self._prop, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def tran(self):
    """ Returns the transitional probability expression if specified """
    return self._tran

  def set_tran(self, tran=None, *args, **kwds):
    """ Sets the conditional transition function with optional arguments and 
    keywords.

    :param tran: may be a scalar, covariance array, or callable function.
    :param *args: optional arguments to pass if tran is callable.
    :param **kwds: optional keywords to pass if tran is callable.
    """
    # Set transition function
    self._tran = tran
    self._sym_tran = False
    self._tsteps = None
    if self._tran is None:
      return
    assert self._anystoch, \
        "Cannot set field transitional without any random variables"
    assert self._prop is None, \
        "Cannot assign both proposition and transition probabilities"
    kwds = dict(kwds)
    if 'tsteps' in kwds:
      self._tsteps = kwds.pop('tsteps')
      if self._tsteps:
        assert type(self._tsteps) is int, "Input tsteps must be int"

    # Pass CF objects directly
    if isinstance(tran, CF):
      assert not args and not kwds, \
        "Setting args and kwds not supported if inputting a CF instance"
      self._tran = CF(self, 
                      tran.ret_inp(), 
                      tran.ret_func(), 
                      *tran.ret_args(), 
                      **tran.ret_kwds())
      self._sym_tran = not self._tran.ret_ismulti()
      return

    # Set as general function
    self._tran = Prob(tran, *args, **kwds)
    self._sym_tran = not self._tran.ismulti

    # If a covariance matrix, set the LU decomposition as the tfun
    if not self._tran.callable and not self._tran.isscalar:
      message = "Non-callable non-scalar tran objects must be a square 2D " + \
                "Numpy array of size corresponding to number of variables {}".\
                 format(self._nvars)
      assert isinstance(tran, np.ndarray), message
      assert tran.ndim == 2, message
      assert np.all(np.array(tran.shape) == self._nvars), message
      self.set_tfun(np.linalg.cholesky(tran))
      assert self._tsteps is None, \
          "Setting tsteps not supported for covariance transitions"
      return

    # If a scipy object, set the tfun
    elif self._tran.isscipy:
      scipyobj = self._tran.expr
      self.set_tfun(self._tran, scipyobj=scipyobj)
      return

    # Return for unit variable RFs
    elif self._unitvar:
      return

    # Otherwise reinstantiate self._tran as an explicit conditional function
    inp = None
    if len(args):
      if isinstance(args[0], (dict, list, tuple)):
        args = tuple(args)
        inp, args = args[0], args[1:]
    self._tran = CF(self, inp, tran, *args, **kwds)
    self._sym_tran = not self._tran.ret_ismulti()

#-------------------------------------------------------------------------------
  @property
  def tfun(self):
    """ Returns the transitional CDF/ICDF expression if specified """
    return self._tfun

  def set_tfun(self, tfun=None, *args, **kwds):
    """ Sets a two-length tuple of functions that should correspond to the
    (cumulative probability function, inverse cumulative function) with respect
    to the callable function set by set_tran(). It is necessary to set these
    functions for conditional sampling variables with non-flat distributions.

    :param tfun: two-length tuple of callable functions or an LU decomposition
    :param *args: arguments to pass to tfun functions
    :param **kwds: keywords to pass to tfun functions
    """
    scipyobj = None if 'scipyobj' not in kwds else kwds['scipyobj']
    self._tfun = tfun 
    if self._tfun is None:
      return
    if 'tsteps' in kwds:
      self._tsteps = kwds.pop('tsteps')

    # Pass CF objects directly
    if isinstance(tfun, CF):
      assert not args and not kwds, \
        "Setting args and kwds not supported if inputting a CF instance"
      self._tfun = CF(self, 
                      tfun.ret_inp(), 
                      tfun.ret_func(), 
                      *tfun.ret_args(), 
                      **tfun.ret_kwds())
      return

    # Handle SciPy objects specifically
    elif scipyobj is not None:      
      lims = np.array([rv.ulims for rv in self._varlist])
      mean = scipyobj.mean
      cov = scipyobj.cov
      self._cond_cov = CondCov(mean, cov, lims)
      scipy_cond = sample_cond_cov if self._sym_tran else \
                   (sample_cond_cov, sample_cond_cov)
      self._tfun = CF(self, None, scipy_cond, cond_cov=self._cond_cov)
      return

    # Non-callables are treated as LUD matrices
    elif not callable(self._tfun): 
        self._tfun = Prob(self._tfun, *args, **kwds)
        message = "Non-callable tran objects must be a triangular 2D Numpy array " + \
                  "of size corresponding to number of variables {}".format(self._nvars)
        assert isinstance(tfun, np.ndarray), message
        assert tfun.ndim == 2, message
        assert np.all(np.array(tfun.shape) == self._nvars), message
        assert np.allclose(tfun, np.tril(tfun)) or \
               np.allclose(tfun, np.triu(tfun)), message
        return

    # Otherwise instantiate a formal conditional function
    inp = None
    if len(args):
      if isinstance(args[0], (dict, list, tuple)):
        args = tuple(args)
        inp, args = args[0], args[1:]
    self._tfun = CF(self, inp, self._tfun, *args, **kwds)

#-------------------------------------------------------------------------------
  def eval_prob(self, values=None, dims=None):
    if values is None:
      values = {}
    else:
      assert isinstance(values, dict), \
          "Input to eval_prob() requires values dict"
      assert set(values.keys()) == self._keyset, \
        "Sample dictionary keys {} mismatch with RV names {}".format(
          values.keys(), self._keylist)

    # If not specified, treat as independent variables
    if self._prob is None or self.__def_prob:
      rvs = self._varlist
      if len(rvs) == 1 and rvs[0]._prob is not None:
        prob = rvs[0].eval_prob(values[rvs[0].name])
      else:
        prob, _ = rv_prod_rule(values, rvs=rvs, pscale=self._pscale)
      return prob

    # Otherwise distinguish between uncallable and callables
    if not self._callable:
      return self._call()
    if self.issympy:
      prob = self._partials['logp'](values) if iscomplex(self._pscale) else \
             self._partials['prob'](values)
      return prob

    # Pass-dims is to replaced when passing Distributions()
    if self._passdims:
      return super().eval_prob(values, dims=dims)
    return super().eval_prob(values)

#-------------------------------------------------------------------------------
  def eval_delta(self, delta=None):
    delta = super().eval_delta(delta)

    # Adjust delta if there is LUD tfun
    if self._tfun is None or self._tfun.isscalar:
      return delta
    elif not self._tfun.callable:
      delta = self._tfun().dot(np.array(delta, dtype=float))
      return self._delta_type(*delta)
    else:
      delta = self._tfun(delta)
      assert isinstance(delta, self._delta_type), \
          "User supplied tfun did not output delta type {}".\
          format(self._delta_type)
      return delta

#-------------------------------------------------------------------------------
  def eval_prop(self, values, **kwargs):
    if self._tran is not None:
      return self.eval_tran(values, **kwargs)
    if values is None:
      values = {}
    if self._prop is None:
      return self.eval_prob(values, **kwargs)
    if not self._prop.callable:
      return self._prop()
    return self._prop(values)

#-------------------------------------------------------------------------------
  def eval_step(self, pred_vals, succ_vals, reverse=False):
    """ Returns adjusted succ_vals """
    if succ_vals is None:
      if self._delta is None:
        if all([isscalar(pred_value) for pred_value in pred_vals]):
          succ_vals = {0}

    # If not sampling succeeding values, use deterministic call
    if not isunitsetint(succ_vals):
      return super().eval_step(pred_vals, succ_vals, reverse=reverse)

    if self._tfun is not None and self._tfun.callable:
      succ_vals = self.eval_tfun(pred_vals)
    elif self._nvars == 1:
      var = self._varlist[0]
      tran = var.tran
      tfun = var.tfun
      if (tran is not None and not tran.callable) or \
          (tfun is not None and tfun.callable):
        vals, dims, kwargs = var.eval_step(pred_vals[var.name], 
                                           succ_vals, reverse=reverse)
        return vals, dims, kwargs
      raise ValueError("Transitional CDF calling requires callable tfun")
    else:
      raise ValueError("Transitional CDF calling requires callable tfun")

    # Initialise outputs with predecessor values
    dims = {}
    kwargs = {'reverse': reverse}
    vals = collections.OrderedDict()
    for key in self._keylist:
      vals.update({key: pred_vals[key]})
    if succ_vals is None and self._tran is None:
      return vals, dims, kwargs

    # If stepping or have a transition function, add successor values
    for key in self._keylist:
      mod_key = key+"'"
      succ_key = key if mod_key not in succ_vals else mod_key
      vals.update({key+"'": succ_vals[succ_key]})

    return vals, dims, kwargs

#-------------------------------------------------------------------------------
  def eval_tfun(self, values, *args, reverse=False, **kwds):
    """ Evaluates tfun from values """

    # Handle non-callables first
    if self._tfun is None:
      return None
    if self._tfun.isscalar:
      if self._tfun.ismulti:
        return self._tfun[int(reverse)]() * values()
      else:
        return self._tfun() * values
    if not self._tfun.callable:
      if self._tfun.ismulti:
        return self._tfun[int(reverse)]().dot(values)
      else:
        return self._tfun().dot(values)

    # If values is not a dictionary then allow a custom call
    if not isinstance(values, dict):
      if self._tfun.ismulti:
        return self._tfun[int(reverse)](values, *args, **kwds).dot(values)
      else:
        return self._tfun(values, *args, **kwds)

    # Support variable sampling one-or-more at-a-time
    succ_vals = collections.OrderedDict(values)

    # For conditional functions, iterate through keys
    assert isinstance(self._tfun, CF), "Callable conditionals must be CF type"

    # Determine keys distinguishing between default and custom specifications
    inp = self._tfun.ret_inp() # ?
    keys = list(succ_vals.keys()) if not inp else list(inp.keys())
    if self._tsteps:
      if self.__cond_mod is None:
        self.__cond_mod = 0
      keys = keys[self.__cond_mod:(self.__cond_mod+self._tsteps)]
      self.__cond_mod += self._tsteps
      if self.__cond_mod >= len(succ_vals):
        self.__cond_mod = 0

    # Handle unspecified input key handling separately
    if not inp:
      if self._tfun.ismulti:
        for key in keys:
          succ_vals[key] = self._tfun[int(reverse)](succ_vals, *args, unknown=key, **kwds)
      else:
        for key in keys:
          succ_vals[key] = self._tfun(succ_vals, *args, unknown=key, **kwds)
      return succ_vals

    # Explicit conditionals require more sophisticated handling
    assert not reverse, \
        "Reverse-direction sampling not supported for explicit conditionals"
    assert self._tfun.ismulti, \
        "Transitional sampling function not specified as a multiple sampler"
    for key in keys:
      values = self._tfun[key](succ_vals, *args, **kwds)
      rv_names = [key] if ',' not in keys else key.split(',')
      if len(rv_names) > 1:
        assert isinstance(values, dict), \
            "Ambiguous output from CF {} for RV {}".format(self._tfun, key)
        for rv_key, rv_val in values.items():
          assert rv_key in succ_vals.keys(), \
              "Key {} not found in {}".format(rv_key, succ_vals.keys())
          succ_vals[key] = rv_val
      else:
        rv_name = rv_names[0]
        if isinstance(values, dict):
          assert rv_name in values.keys(), \
              "Value key {} not found from output {}".format(rv_name, values.keys())
          succ_vals[rv_name] = values[rv_name]
        else:
          succ_vals[rv_name] = values
    return succ_vals

#-------------------------------------------------------------------------------
  def eval_tran(self, values, **kwargs):
    if 'cond' in kwargs:
      return kwargs['cond']
    cond = DEFAULT_CONDITIONAL_PROBABILITY[iscomplex(self._pscale)]
    reverse = False if 'reverse' not in kwargs else kwargs['reverse']
    if self._tran is None:
      if self._tfun is not None: # tfun without tran means no cond. prob.
        return cond
      rvs = self._varlist
      if len(rvs) == 1 and rvs[0]._tran is not None:
        return rvs[0].eval_tran(values, **kwargs)
      pred_vals = dict()
      succ_vals = dict()
      for key_, val in values.items():
        prime = key_[-1] == "'"
        key = key_[:-1] if prime else key_
        if key in self._keylist:
          if prime:
            succ_vals.update({key: val})
          else:
            pred_vals.update({key: val})
      cond, _ = rv_prod_rule(pred_vals, succ_vals, rvs=rvs, pscale=self._pscale)
    elif not self._tran.callable or self._tran.isscipy:
      return cond
    else:
      cond = self._tran(values) if self._sym_tran else \
             self._tran[int(reverse)](values)
    return cond

#-------------------------------------------------------------------------------
  def reval_tran(self, dist):
    """ Evaluates the conditional reverse-transition function for corresponding 
    transition conditional distribution dist. This requires a tuple input for
    self.set_tran() to evaluate a new conditional.
    """
    assert isinstance(dist, PD), \
        "Input must be a distribution, not {} type.".format(type(dist))
    marg, cond = dist.cond, dist.marg
    name = margcond_str(marg, cond)
    vals = collections.OrderedDict(dist)
    dims = dist.dims
    # This next line needs to be modified to handle new API
    """
    prob = dist.prob if self._sym_tran or self._tran is None \
           else self._tran[1](dist)
    """
    prob = dist.prob
    pscale = dist.pscale
    return PD(name, vals, dims=dims, prob=prob, pscale=pscale)

#-------------------------------------------------------------------------------
  def _eval_iid(self, dist_name, vals, dims, prob, iid):
    if not iid: 
      return PD(dist_name, vals, dims=dims, prob=prob, pscale=self._pscale)

    # Deal with IID cases
    max_dim = None
    for dim in dims.values():
      if dim is not None:
        max_dim = dim if max_dim is None else max(dim, max_dim)

    # If scalar or prob is expected shape then perform product here
    if max_dim is None or max_dim == prob.ndim - 1:
      dist = PD(dist_name, vals, dims=dims, prob=prob, pscale=self._pscale)
      return dist.prod(iid)

    # Otherwise it is left to the user function to perform the iid product
    for key in iid:
      vals[key] = {len(vals[key])}
      dims[key] = None

    # Tidy up probability
    return PD(dist_name, vals, dims=dims, prob=prob, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ Returns a joint distribution p(args) """
    if not self.anystoch:
      return Field.__call__(self, *args, **kwds)
    if not self._nvars:
      return None
    iid = False if 'iid' not in kwds else kwds.pop('iid')
    if type(iid) is bool and iid:
      iid = self._defiid
    if not kwds and len(args) == 1 and not isinstance(args[0], dict):
      arg = {key: args[0] for key in self._keyset}
      args = arg,
    values = self.parse_args(*args, **kwds)
    dist_name = self.eval_dist_name(values)
    vals, dims = self.evaluate(values, _skip_parsing=True)
    prob = self.eval_prob(vals, dims)
    return self._eval_iid(dist_name, vals, dims, prob, iid)

#-------------------------------------------------------------------------------
  def propose(self, *args, **kwds):
    """ Returns a proposal distribution p(args[0]) for values """
    suffix = "'" if 'suffix' not in kwds else kwds.pop('suffix')
    if not kwds and len(args) == 1 and not isinstance(args[0], dict):
      arg = {key: args[0] for key in self._keyset}
      args = arg,
    values = self.parse_args(*args, **kwds)
    dist_name = self.eval_dist_name(values, suffix)
    vals, dims = self.evaluate(values, _skip_parsing=True)
    prop = self.eval_prop(vals) if self._prop is not None else \
           self.eval_prob(vals, dims)
    if suffix:
      keys = list(vals.keys())
      for key in keys:
        mod_key = key + suffix
        vals.update({mod_key: vals.pop(key)})
        if key in dims:
          dims.update({mod_key: dims.pop(key)})
    return PD(dist_name, vals, dims=dims, prob=prop, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def step(self, *args, **kwds):
    """ Returns a proposal distribution p(args[1]) given args[0], depending on
    whether using self._prop, that denotes a simple proposal distribution,
    or self._tran, that denotes a transitional distirbution. """

    reverse = False if 'reverse' not in kwds else kwds.pop('reverse')
    pred_vals, succ_vals = None, None 
    if len(args) == 1:
      if isinstance(args[0], (list, tuple)) and len(args[0]) == 2:
        pred_vals, succ_vals = args[0][0], args[0][1]
      else:
        pred_vals = args[0]
    elif len(args) == 2:
      pred_vals, succ_vals = args[0], args[1]

    # Evaluate predecessor values
    if not isinstance(pred_vals, dict):
      pred_vals = {key: pred_vals for key in self._keyset}
    pred_vals = self.parse_args(pred_vals, pass_all=True)
    dist_pred_name = self.eval_dist_name(pred_vals)
    pred_vals, pred_dims = self.evaluate(pred_vals)

    # Default successor values if None and delta is None
    if succ_vals is None and self._delta is None:
      pred_values = list(pred_vals.values())
      if all([isscalar(pred_value) for pred_value in pred_values]):
        succ_vals = {0}
      else:
        succ_vals = pred_vals

    # Evaluate successor evaluates
    vals, dims, kwargs = self.eval_step(pred_vals, succ_vals, reverse=reverse)
    succ_vals = {key[:-1]: val for key, val in vals.items() if key[-1] == "'"}
    cond = self.eval_tran(vals, **kwargs)
    dist_succ_name = self.eval_dist_name(succ_vals, "'")
    dist_name = '|'.join([dist_succ_name, dist_pred_name])

    return PD(dist_name, vals, dims=dims, prob=cond, pscale=self._pscale)

#-------------------------------------------------------------------------------
  def __eq__(self, other):
    """ Equality for RFs is defined as comprising the same RVs """
    if type(self) is not RF or type(other) is not RF:
      return super().__eq__(other)
    return self._keyset == other.keyset

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    if not self.issympy or isinstance(arg, int) or \
        (isinstance(arg, str) and arg in self._keyset):
      return Field.__getitem__(self, arg)
    return Prob.__getitem__(self, arg)

#-------------------------------------------------------------------------------



""" Provides utility functions for Functional """
import collections
import networkx as nx

#-------------------------------------------------------------------------------
NX_UNDIRECTED_GRAPH = nx.Graph

#-------------------------------------------------------------------------------
def collate_vertices(graph=None):
  """ Processes through all nodes of graph returning a dictionary in the form
  {node_name: node_instance}
  """
  vertices = collections.OrderedDict()
  if not graph:
    return vertices
  if not isinstance(graph, NX_UNDIRECTED_GRAPH):
    raise TypeError("Expecting type {} but inputted {}".format(
      NX_UNDIRECTED_GRAPH, type(graph)))
  for vertex in list(graph.nodes):
    assert hasattr(vertex, 'name'), "Every vertex must include name attribute"
    vertex_name = vertex.name
    assert isinstance(vertex_name, str), "Vertex name must be a string"
    assert vertex_name.isidentifier(), "Vertex name must be an identifier"
    vertices.update({vertex_name: vertex})
  return vertices

#-------------------------------------------------------------------------------
def parse_identifiers(identifiers):
  """ Reads identifiers, which may be a string or list/tuple/set of objects 
  instances with name instances as string, returning a frozen set of names.
  """
  if isinstance(identifiers, str):
    return frozenset(identifiers.split(','))
  if not isinstance(identifiers, (list, tuple, set)):
    identifiers = identifiers,
  keys = list(identifiers)
  for i, key in enumerate(keys):
    if not isinstance(key, str):
      assert hasattr(key, 'name'), \
        "Each element in hashable tuple must be a string or have name attribute"
      key_name = key.name
      assert isinstance(key_name, str), \
        "Each non-string hashable tuple element must a string name attribute"
      keys[i] = key_name
  return frozenset(keys)

#-------------------------------------------------------------------------------



"""
A field is a collection of a functionally-related variables without explictly
directional dependences.
"""
#-------------------------------------------------------------------------------
import collections
import numpy as np

from probayes.variable import Variable
from probayes.variable_utils import parse_as_str_dict
from probayes.vtypes import isscalar, isunitsetint, issingleton, isdimensionless
from probayes.pscales import real_sqrt
from probayes.ops import and_op, or_op
from probayes.expression import Expression
from probayes.functional import NX_UNDIRECTED_GRAPH
from probayes.distribution import Distribution

#-------------------------------------------------------------------------------
class Field (NX_UNDIRECTED_GRAPH):
  """
  A field is a collection of a variables that are functionally inter-related
  without explicit directional dependencies. This lack of directionality means
  Field does not support Functionals. 
  """

  # Protected
  _name = None       # Random field name cannot be set externally
  _Delta = None      # A namedtuple generator for delta operations
  _vars = None       # OrderedDict of variables
  _nvars = None      # Number of variables
  _unitvar = None    # Boolean flag for one variable
  _anyfloat = None   # Boolean flag to denote if any variables are float type
  _anystoch = None   # Boolean flag of whether any variable stochastic
  _varlist = None    # List of variable objects
  _keylist = None    # List of keys of variable names
  _keyset = None     # Unordered set of keys of random variable names
  _delta = None      # Delta function (to replace step)
  _delta_args = None # Optional delta args (must be dictionaries)
  _delta_kwds = None # Optional delta kwds
  _delta_type = None # Proxy for delta used for casting
  _length = None     # Length of field
  _lengths = None    # Lengths of Variables
  _leafs = None      # Field of Variables that do not condition others (for SD)
  _roots = None      # Field of Vairables not dependent on others (for SD)
  _stems = None      # OrderedDict of latent Variables (for Dependencies)
  _is_stochastic = False      # Flag of whether stochastic
  _func = False      # Functional

#-------------------------------------------------------------------------------
  @property
  def is_stochastic(self):
    return self._is_stochastic

  def __init__(self, *args): # over-rides NX_GRAPH.__init__()
    """ Initialises a random field with Variables for in args. See set_vars(). """
    super().__init__()
    self.set_vars(*args)
    self.set_delta()

#-------------------------------------------------------------------------------
  def add_node(self, *args, **kwds):
    """ Direct adding of nodes disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  def add_node_from(self, *args, **kwds):
    """ Direct adding of nodes disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  def remove_node(self, *args, **kwds):
    """ Removal of nodes disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  def add_edge(self, *args, **kwds):
    """ Direct adding of edges disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  def add_edges_from(self, *args, **kwds):
    """ Direct adding of edges disallowed """
    raise NotImplementedError("Not implemented for this class")

#-------------------------------------------------------------------------------
  @property
  def leafs(self):
    return self._leafs

  @property
  def stems(self):
    return self._stems

  @property
  def roots(self):
    return self._roots

  def set_vars(self, *args):
    """ Initialises a field with variables for each arg in args.

    :param *args: each arg may be an Variable instance or the first arg may be a Field.
    
    """
    self._leafs, self._stems, self._roots = self, None, None
    if len(args) == 1 and isinstance(args[0], (Variable, set, tuple, list)):
      args = args[0]
    else:
      args = tuple(args)
    self.add_var(args)

#-------------------------------------------------------------------------------
  def add_var(self, var):
    """ Adds one or more variables to the field.

    :param var: a Variable or Field instance, or a list/tuple of Variable instances.
    """
    variables = []
    if isinstance(var, (Field, dict, set, tuple, list)):
      variables = var
      if isinstance(variables, Field):
        variables = variables.varlist
      elif isinstance(variables, dict):
        variables = list(variables.values())
    for var in variables:
      assert isinstance(var, Variable), \
          "Input not a variable instance but of type: {}".format(type(var))
      if self._nvars:
        assert var not in list(self.nodes), \
            "Existing variable {} already present in collection".format(var)
      super().add_node(var)
    else:
      assert isinstance(var, Variable), \
          "Input not a variable instance but of type: {}".format(type(var))
      if self._nvars:
        assert var not in list(self.nodes), \
            "Existing variable {} already present in collection".format(var)
      super().add_node(var)
    self._refresh()

#-------------------------------------------------------------------------------
  @property
  def vars(self):
    return self._vars

  @property
  def nvars(self):
    return self._nvars

  @property
  def unitvar(self):
    return self._unitvar

  @property
  def anyfloat(self):
    return self._anyfloat

  @property
  def anystoch(self):
    return self._anystoch

  def __len__(self):
    return self._nvars

  def __bool__(self):
    if self._nvars:
      return True
    return False

  @property
  def varlist(self):
    return self._varlist

  @property
  def keylist(self):
    return self._keylist

  @property
  def keyset(self):
    return self._keyset

  @property
  def name(self):
    return self._name

  @property
  def id(self):
    return self._id

  @property
  def Delta(self):
    return self._Delta

  def _refresh(self):
    """ Updates Variable summary objects, Field name and id, and delta factory. """
    self._vars = collections.OrderedDict({var.name: var for var in list(self.nodes)})
    self._nvars = self.number_of_nodes()
    self._unitvar = self._nvars == 1
    self._varlist = list(self._vars.values())
    self._anyfloat = any([var.vtype is float for var in self._varlist])
    self._anystoch = any([var.is_stochastic for var in self._varlist])
    if not self._is_stochastic:
      assert not self._anystoch,\
          "Stochastic variables cannot contribute to a non-stochastic field"
    self._keylist = list(self._vars.keys())
    self._keyset = frozenset(self._keylist)
    self._defiid = self._leafs.keylist
    self._name = ','.join(self._keylist)
    self._id = '_and_'.join(self._keylist)
    if self._id:
      self._Delta = collections.namedtuple('ð', self._keylist)
      self._delta_type = self._Delta
    self.eval_length()

#-------------------------------------------------------------------------------
  @property
  def delta(self):
    """ Returns the default delta object if specified """
    return self._delta

  def set_delta(self, delta=None, *args, **kwds):
    """ Sets the default delta function or operation.

    :param delta: the delta function or operation (see below)
    :param *args: optional arguments to pass if delta is callable.
    :param **kwds: optional keywords to pass if delta is callable.

    The input delta may be:

    1. A callable function (for which args and kwds are passed on as usual).
    2. An Variable.Delta instance (this defaults all Variable Deltas).
    3. A dictionary for Variables, this is converted to an Field.Delta.
    4. A scalar that may contained in a list or tuple:
      a) No container - the scalar is treated as a fixed delta.
      b) List - delta is uniformly and independently sampled across Variabless.
      c) Tuple - delta is spherically sampled across Variables.

      For non-tuples, an optional argument (args[0]) can be included as a 
      dictionary to specify by Variable-name deltas following the above 
      conventions except their values are not subject to scaling even if 
      'scale' is given, but they are subject to bounding if 'bound' is 
      specified.

    For setting types 2-4, optional keywords are (default False):
      'scale': Flag to denote scaling deltas to Variable lengths
      'bound': Flag to constrain delta effects to Variable bounds (None bounces)
      
    """
    self._delta = delta
    self._delta_args = args
    self._delta_kwds = dict(kwds)
    self._spherise = {}
    if self._delta is None:
      return
    elif callable(self._delta):
      self._delta = Expression(self._delta, *args, **kwds)
      return

    # Default scale and bound
    if 'scale' not in self._delta_kwds:
      self._delta_kwds.update({'scale': False})
    if 'bound' not in self._delta_kwds:
      self._delta_kwds.update({'bound': False})
    scale = self._delta_kwds['scale']
    bound = self._delta_kwds['bound']

    # Handle deltas and dictionaries
    if isinstance(self._delta, dict):
      self._delta = self._delta_type(**self._delta)
    if isinstance(delta, self._delta_type):
      assert not args, \
        "Optional args prohibited for dict/delta instance inputs"
      for i, var in enumerate(self._varlist):
        var.set_delta(self._delta[i], scale=scale, bound=bound)
      return

    # Default scale and bound and check args
    if self._delta_args:
      assert len(self._delta_args) == 1, \
          "Optional positional arguments must comprises a single dict"
      unscale = self._delta_args[0]
      assert isinstance(unscale, dict), \
          "Optional positional arguments must comprises a single dict"

    # Non tuples can be converted to deltas; can pre-scale here
    if not isinstance(self._delta, tuple):
      delta = self._delta 
      urand = isinstance(delta, list)
      if urand:
        assert len(delta) == 1, "List delta requires a single element"
        delta = delta[0]
      deltas = {key: delta for key in self._keylist}
      unscale = {} if not self._delta_args else self._delta_args[0]
      deltas.update(unscale)
      delta_dict = collections.OrderedDict(deltas)
      for i, (key, val) in enumerate(deltas.items()):
        delta = val
        if scale and key not in unscale:
          assert np.isfinite(self._lengths[i]), \
              "Cannot scale by infinite length for Variable {}".format(key)
          delta = val * self._lengths[i]
        if urand:
          delta = [delta]
        delta_dict.update({key: delta})
      self._delta = self._delta_type(**delta_dict)
      for i, var in enumerate(self._varlist):
        var.set_delta(self._delta[i], scale=False, bound=bound)

    # Tuple deltas must be evaluated on-the-fly and cannot be pre-scaled
    else:
      unscale = {} if not self._delta_args else self._delta_args[0]
      self._spherise = {}
      for i, key in enumerate(self._keylist):
        if key not in unscale.keys():
          length = self._lengths[i]
          assert np.isfinite(length), \
              "Cannot spherise Variable {} with infinite length".format(key)
          self._spherise.update({key: length})

#-------------------------------------------------------------------------------
  def eval_length(self):
    """ Evaluates and returns the joint length of the field. """
    self._lengths = np.array([var.length for var in self._varlist], 
                             dtype=float)
    self._length = np.sqrt(np.sum(self._lengths**2))
    return self._length

#-------------------------------------------------------------------------------
  def parse_args(self, *args, **kwds):
    """ Returns (values, iid) from *args and **kwds """
    pass_all = False if 'pass_all' not in kwds else kwds.pop('pass_all')
    values = parse_as_str_dict(*args, **kwds)
    seen_keys = []
    for key, val in values.items():
      count_comma = key.count(',')
      if count_comma:
        seen_keys.extend(key.split(','))
        if isinstance(val, (tuple, list)):
          assert len(val) == count_comma+1, \
              "Mismatch in key specification {} and number of values {}".\
              format(key, len(val))
        else:
          values.update({key: [val] * (count_comma+1)})
      else:
        seen_keys.append(key)
      if not pass_all:
        assert seen_keys[-1] in self._keyset, \
            "Unrecognised key {} among available Variables {}".format(
                seen_keys[-1], self._keyset)

    # Default values
    def_val = None
    if self._anyfloat:
      if not len(values):
        def_val = {0}
      else:
        def_val = {0}
        for val in values.values():
          if isunitsetint(val):
            if val == {0}:
              def_val = None
    for key in self._keylist:
      if key not in seen_keys:
        values.update({key: def_val})
    if pass_all:
      list_keys = list(values.keys())
      for key in list_keys:
        if key not in self._keylist:
          values.pop(key)

    return values

#-------------------------------------------------------------------------------
  def eval_dist_name(self, values=None, suffix=None):
    # Evaluates the string used to set the distribution name
    vals = collections.OrderedDict()
    if isinstance(values, dict):
      for key, val in values.items():
        if ',' in key:
          subkeys = key.split(',')
          for i, subkey in enumerate(subkeys):
            vals.update({subkey: val[i]})
        else:
          vals.update({key: val})
      for key in self._keylist:
        if key not in vals.keys():
          vals.update({key: None})
    else:
      vals.update({key: values for key in self._keylist})
    var_dist_names = [var.eval_dist_name(vals[var.name], suffix) \
                     for var in self._varlist]
    dist_name = ','.join(var_dist_names)
    return dist_name

#-------------------------------------------------------------------------------
  def evaluate(self, *args, _skip_parsing=False, min_dim=0, **kwds):
    """ 
    Keep args and kwds since could be called externally. This ignores self._prob.
    """
    values = self.parse_args(*args, **kwds) if not _skip_parsing else args[0]
    dims = collections.OrderedDict()
    
    # Don't reshape if all scalars (and therefore by definition no shared keys)
    if all([np.isscalar(value) for value in values.values()]): # use np.scalar
      return values, dims

    # Create reference mapping for shared keys across vars
    values_ref = collections.OrderedDict({key: [key, None] for key in self._keylist})
    for key in values.keys():
      if ',' in key:
        subkeys = key.split(',')
        for i, subkey in enumerate(subkeys):
          values_ref[subkey] = [key, i]

    # Share dimensions for joint variables and do not dimensionalise scalars
    ndim = min_dim
    dims = collections.OrderedDict({key: None for key in self._keylist})
    seen_keys = set()
    for i, key in enumerate(self._keylist):
      new_dim = False
      if values_ref[key][1] is None: # i.e. not shared
        if not isdimensionless(values[key]):
          dims[key] = ndim
          new_dim = True
        seen_keys.add(key)
      elif key not in seen_keys:
        val_ref = values_ref[key]
        subkeys = val_ref[0].split(',')
        for subkey in subkeys:
          dims[subkey] = ndim
          seen_keys.add(subkey)
        if not isdimensionless(values[val_ref[0]][val_ref[1]]):
          new_dim = True
      if new_dim:
        ndim += 1

    # Reshape
    vdims = [dim for dim in dims.values() if dim is not None]
    ndims = max(vdims) + 1 if len(vdims) else 0
    ones_ndims = np.ones(ndims, dtype=int)
    vals = collections.OrderedDict()
    for i, var in enumerate(self._varlist):
      key = var.name
      reshape = True
      if key in values.keys():
        vals.update({key: values[key]})
        reshape = not np.isscalar(vals[key])
        if vals[key] is None or isinstance(vals[key], set):
          vals.update(var.evaluate(vals[key]))
      else:
        val_ref = values_ref[key]
        vals_val = values[val_ref[0]][val_ref[1]]
        if vals_val is None or isinstance(vals_val, set):
          vals.update(var.evaluate(vals_val))
        else:
          vals.update({key: vals_val})
      if reshape and not isscalar(vals[key]):
        re_shape = np.copy(ones_ndims)
        re_dim = dims[key]
        re_shape[re_dim] = vals[key].size
        vals[key] = vals[key].reshape(re_shape)
    
    # Remove dimensionality for singletons
    for key in self._keylist:
      if issingleton(vals[key]):
        dims[key] = None
    return vals, dims

#-------------------------------------------------------------------------------
  def eval_delta(self, delta=None):

    # Handle native delta types within Variable deltas
    if delta is None: 
      if self._delta is None:
        variables = list(self._varlist)
        if len(variables) == 1 and variables[0]._delta is not None:
          return variables[0]._delta
        return None
      elif isinstance(self._delta, Expression):
        delta = self._delta()
      elif isinstance(self._delta, self._delta_type):
        delta_dict = collections.OrderedDict()
        variables = self._varlist
        for i, key in enumerate(self._keylist):
          delta_dict.update({key: variables[i].eval_delta()})
        delta = self._delta_type(**delta_dict)
      else:
        delta = self._delta
    elif isinstance(delta, Expression):
      delta = delta()
    elif isinstance(delta, self._delta_type):
      delta_dict = collections.OrderedDict()
      variables = self.ret_vars(aslist=True)
      for i, key in enumerate(self._keylist):
        delta_dict.update({key: variables[i].eval_delta(delta[i])})
      delta = self._delta_type(**delta_dict)

    # Non spherical case
    if not isinstance(self._delta_type, Expression) and \
         isinstance(delta, self._delta_type): # i.e. non-spherical
      return delta

    # Rule out possibility of all variables contained in unscaling argument
    assert isinstance(delta, tuple), \
        "Unknown delta type: {}".format(delta)
    unscale = {} if not self._delta_args else self._delta_args
    if not len(self._spherise):
      return self._delta_type(**unscale)

    # Spherical version
    delta = delta[0]
    spherise = self._spherise
    keys = self._spherise.keys()
    rss = real_sqrt(np.sum(np.array(list(spherise.values()))**2))
    if self._delta_kwds['scale']:
      delta *= rss
    deltas = np.random.uniform(-delta, delta, size=len(spherise))
    rss_deltas = real_sqrt(np.sum(deltas ** 2.))
    deltas = (deltas * delta) / rss_deltas
    delta_dict = collections.OrderedDict()
    idx = 0
    for i, key in enumerate(keys):
      if key in unscale:
        val = unscale[key]
      else:
        val = deltas[idx]
        idx += 1
        if self._delta_kwds['scale']:
          val *= self._lengths[i]
      delta_dict.update({key: val})
    delta = self._delta_type(**delta_dict)
    return delta

#-------------------------------------------------------------------------------
  def apply_delta(self, values, delta=None):
    delta = delta or self._delta
    if delta is None:
      return values
    if not isinstance(delta, self._delta_type):
      variables = self._varlist
      if len(variables) == 1 and isinstance(delta, variables[0].delta):
        return variables[0].apply_delta(values, delta)
      raise TypeError("Cannot apply delta without providing delta type {}".\
        format(self._delta_type))
    bound = False if 'bound' not in self._delta_kwds \
           else self._delta_kwds['bound']
    vals = collections.OrderedDict(values)
    keys = delta._fields
    for i, key in enumerate(keys):
      vals.update({key: self._vars[key].apply_delta(values[key], 
                                                    delta[i], 
                                                    bound=bound)})
    return vals

#-------------------------------------------------------------------------------
  def eval_step(self, pred_vals, succ_vals, reverse=False):
    """ Returns adjusted succ_vals """

    # Evaluate deltas if required
    if succ_vals is None:
      if self._delta is None:
        pred_values = list(pred_vals.values())
        if all([isscalar(pred_value) for pred_value in pred_values]):
          raise ValueError("Stochastic step sampling not supported for Field; use RF")
        else:
          succ_vals = pred_vals
      else:
        succ_vals = self.eval_delta()
    elif isinstance(succ_vals, Expression) or \
        isinstance(succ_vals, (tuple, self._delta_type)):
      succ_vals = self.eval_delta(succ_vals)

    # Apply deltas
    cond = None
    if isinstance(succ_vals, self._delta_type):
      succ_vals = self.apply_delta(pred_vals, succ_vals)
    elif isunitsetint(succ_vals):
      raise ValueError("Stochastic step sampling not supported for Field; use RF")

    # Initialise outputs with predecessor values
    dims = {}
    kwargs = {'reverse': reverse}
    if cond is not None:
      kwargs = {'cond': cond}
    vals = collections.OrderedDict()
    for key in self._keylist:
      vals.update({key: pred_vals[key]})
    if succ_vals is None:
      return vals, dims, kwargs

    # If stepping, add successor values
    for key in self._keylist:
      mod_key = key+"'"
      succ_key = key if mod_key not in succ_vals else mod_key
      vals.update({key+"'": succ_vals[succ_key]})

    return vals, dims, kwargs

#-------------------------------------------------------------------------------
  def subfield(self, vertices):
    """ Returns a view of vertices, which must all be members, as a Field.

    :param vertices: str/variable/field or list/tuple of str/variable/field use.

    :return: an Field including only those vertices
    """

    # Convert to list
    if isinstance(vertices, tuple):
      vertices = list(vertices)
    if not isinstance(vertices, list):
      vertices = [vertices]

    # Collate variables and return as Field
    variables = []
    for vertex in vertices:
      if isinstance(vertex, Field):
        variables += [vertex.ret_vars(aslist=True)]
      elif isinstance(vertex, Variable):
        variables += [vertex]
      elif isinstance(vertex, str):
        variables += [self._vars[vertex]]
      else:
        raise TypeError("Unrecognised vertex specification type: {}".format(
            type(vertex)))
    return Field(*tuple(variables))

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ Returns a distribution p(args) """
    if not self._nvars:
      return None
    iid = False if 'iid' not in kwds else kwds.pop('iid')
    if type(iid) is bool and iid:
      iid = self._defiid
    values = self.parse_args(*args, **kwds)
    vals, dims = self.evaluate(values, _skip_parsing=True)
    name = ','.join(list(vals.keys()))
    return Distribution(name, vals, dims=dims)

#-------------------------------------------------------------------------------
  def __and__(self, other):
    return and_op(self, other)

#-------------------------------------------------------------------------------
  def __or__(self, other):
    return or_op(self, other)

#-------------------------------------------------------------------------------
  def __eq__(self, other):
    """ Equality for Fields is defined as comprising the same Variables """
    if type(self) is not Field or type(other) is not Field:
      return super().__eq__(other)
    return self._keyset == other.keyset

#-------------------------------------------------------------------------------
  def __ne__(self, other):
    """ Equality for Field is defined as comprising the same Variables """
    return not self.__eq__(other)

#-------------------------------------------------------------------------------
  def __getitem__(self, key):
    if type(key) is int:
      key = self._keylist[key]
    if isinstance(key, str):
      if key not in self._keylist:
        return None
    return self._vars[key]

#-------------------------------------------------------------------------------
  def __repr__(self):
    if not self._name:
      return NX_UNDIRECTED_GRAPH.__repr__(self)
    return NX_UNDIRECTED_GRAPH.__repr__(self) + ": '" + self._name + "'"

#-------------------------------------------------------------------------------



""" Provides SP, a class for stochastic processes """

#-------------------------------------------------------------------------------
import collections
import inspect
import warnings
from probayes.sd import SD
from probayes.expression import Expression
from probayes.pd import PD
from probayes.pd_utils import summate
from probayes.sp_utils import sample_generator, MCMC_SAMPLERS

#-------------------------------------------------------------------------------
class SP (SD):
  """ A stocastic process is stochastic dependence that supports an indexable 
  sequence of realisations from a directed graph. It is therefore implemented 
  here using a sample generator that iteratively samples a directed graph.
  """

  # Protected
  _scores = None # Scores function used for the basis of acceptance
  _thresh = None # Threshold function to compare with scores
  _update = None # Update function (output True, None, or False)

  # Private
  __samplers = None  # List of samplers
  __counter = None   # Step counter
  __last = None      # Last argument ordereddict

#-------------------------------------------------------------------------------
  def __init__(self, *args, **kwds):
    super().__init__(*args, **kwds)
    self.reset()

#-------------------------------------------------------------------------------
  @property
  def stuv(self):
    return self._stuv

  @property
  def opqrstuv(self):
    return self._opqrstuv

  def _refresh(self, *args, **kwds):
    super()._refresh(*args, **kwds)
    if not self._nvars:
      return
    self._stuv = collections.namedtuple(self._id, ['s', 't', 'u', 'v'])
    self._opqrstuv = collections.namedtuple(self._id, 
                        ['o', 'p', 'q', 'r', 's', 't', 'u', 'v'])

#-------------------------------------------------------------------------------
  @property
  def scores(self):
    return self._scores

  def set_scores(self, scores=None, *args, **kwds):
    self._scores = scores
    if self._scores is None:
      return
    if self._scores in MCMC_SAMPLERS:
      assert not args and not kwds, \
          "Neither args nor kwds permitted with spec '{}'".format(self._scores)
      self.set_scores(MCMC_SAMPLERS[self._scores][0], pscale=self._pscale)
      self.set_thresh(scores)
      return
    self._scores = Expression(self._scores, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def thresh(self):
    return self._thresh

  def set_thresh(self, thresh=None, *args, **kwds):
    self._thresh = thresh
    if self._thresh is None:
      return
    if self._thresh in MCMC_SAMPLERS:
      assert not args and not kwds, \
          "Neither args nor kwds permitted with spec '{}'".format(self._thresh)
      self.set_thresh(MCMC_SAMPLERS[self._thresh][1])
      self.set_update(thresh)
      return
    self._thresh = Expression(self._thresh, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def update(self):
    return self._update

  def set_update(self, update=None, *args, **kwds):
    self._update = update
    if self._update is None:
      return
    if self._update in MCMC_SAMPLERS:
      assert not args and not kwds, \
          "Neither args nor kwds permitted with spec '{}'".format(self._update)
      self.set_update(MCMC_SAMPLERS[self._update][2])
      return
    self._update = Expression(self._update, *args, **kwds)

#-------------------------------------------------------------------------------
  def eval_expr(self, expr=None, *args, **kwds):
    if expr is None:
      return expr
    assert isinstance(expr, Expression), \
        "Evaluation of expr no possible for type {}".format(type(expr))
    if not expr.callable:
      return expr()
    return expr(*args, **kwds)

#-------------------------------------------------------------------------------
  def reset(self, sampler_id=None, reset_last=True): # Leave option to preseve
    if self.__samplers is None or sampler_id is None:
      self.__samplers = []
    if self.__counter is None:
      self.__counter = collections.Counter()
    if self.__last is None:
      self.__last = collections.OrderedDict()
    if sampler_id is None:
      return
    sampler = self.get_sampler(sampler_id)
    self.__counter[sampler] = 0
    if sampler not in self.__last:
      self.__last[sampler] = None
    elif reset_last:
      self.__last[sampler] = None
    return self.__samplers, self.__counter, self.__last

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    conditionalise = None if 'conditionalise' not in kwds else \
                     kwds.pop('conditionalise')

    # Summating distributions is straightforward
    samples = None if not len(args) else args[0]
    if samples and isinstance(samples, (list, tuple, collections.deque)) \
        and len(samples) and isinstance(samples[0], PD):
      samples = list(samples)
      for sample in samples[1:]:
        assert isinstance(sample, PD),\
            "If using distributions, all samples must be distributions, not {}".\
            format(type(sample))
      if isinstance(samples, (list, collections.deque)):
        samples = tuple(samples)
      dist = summate(*samples)
      if not conditionalise:
        return dist
      return dist.conditionalise(self._leafs.keyset)

    # If empty call or dictionary arguments, pass to SD
    if not samples or len(kwds) or \
        not isinstance(samples, (list, tuple, collections.deque)):
      return super().__call__(*args, **kwds)
    if not len(samples):
      return super().__call__(*args, **kwds)

    # Otherwise this call is SP-specific
    opqrstuv = collections.OrderedDict({key: None for key in \
        ['o', 'p', 'q', 'r', 's', 't', 'u', 'v']})

    # We exclude update=False from the summation
    opqrstuv['u'] = []

    def _maybe_append(element, key):
      if element is not None:
        if opqrstuv[key] is None:
          opqrstuv[key] = []
        opqrstuv[key].append(element)

    for sample in samples:
      if isinstance(sample, self.opqr):
        _maybe_append(sample.o, 'o')
        _maybe_append(sample.p, 'p')
        _maybe_append(sample.q, 'q')
        _maybe_append(sample.r, 'r')
      
      else:
        assert isinstance(sample, self._opqrstuv), \
            "Sample must be outputted from sampler: {}".format(self._id)
        if sample.u == False:
          continue
        if self._update is not None:
          opqrstuv['u'].append(sample.u)
        _maybe_append(sample.o, 'o')
        _maybe_append(sample.p, 'p')
        _maybe_append(sample.q, 'q')
        _maybe_append(sample.r, 'r')
        _maybe_append(sample.s, 's')
        _maybe_append(sample.t, 't')
        _maybe_append(sample.v, 'v')
          
    for key in ['o', 'p', 'q', 'r', 'v']:
      if opqrstuv[key] is not None:
        opqrstuv[key] = summate(*tuple(opqrstuv[key]))
        if conditionalise and key in ['o', 'p', 'v']:
          opqrstuv[key] = opqrstuv[key].conditionalise(self._leafs.keyset)
    return self._opqrstuv(**opqrstuv)

#-------------------------------------------------------------------------------
  def get_sampler(self, sampler_id=None):
    if sampler_id is None:
      return self.__samplers
    if type(sampler_id) is int:
      return self.__samplers[sampler_id]
    return sampler_id

#-------------------------------------------------------------------------------
  def get_counter(self, sampler_id=None):
    if sampler_id is None:
      return self.__counter
    return self.__counter[self.get_sampler(sampler_id)]

#-------------------------------------------------------------------------------
  def get_last(self, sampler_id=None):
    if sampler_id is None:
      return self.__last
    return self.__last[self.get_sampler(sampler_id)]

#-------------------------------------------------------------------------------
  def next(self, sampler_id, *args, **kwds):

    # Reset counters
    sampler = self.get_sampler(sampler_id)
    self.__counter[sampler] += 1
    last = self.__last[sampler]
    no_proposal = self._tran is None and self._tfun is None and \
                  not self._unit_tran and self._prop is None

    # Treat sampling without proposals as a distribution call
    if last is None or no_proposal:
      opqr = self.sample(*args, **kwds)
      if no_proposal:
        return opqr

    # Otherwise refeed last proposals into sample function
    else:
      if self._tran is None and self._tfun is None and not self._unit_tran and \
          self._delta is None:
        last = {0}
      if len(args) < 2:
        opqr = self.sample(last, **kwds)
      else:
        args = tuple([last] + list(args[1:]))
        opqr = self.sample(*args, **kwds)

    # Set to last if accept is not False
    stuv = self._stuv(self.eval_expr(self._scores, opqr),
                      self.eval_expr(self._thresh),
                      None,
                      None)
    update = self.eval_expr(self._update, stuv)
    verdit = opqr.o
    if self._update is None or self.__last[sampler] is None or update:
      self.__last[sampler] = opqr
      verdit = opqr.p
    return self._opqrstuv(opqr.o, opqr.p, opqr.q, opqr.r, 
                          stuv.s, stuv.t, update, verdit)

#-------------------------------------------------------------------------------
  def sampler(self, *args, **kwds):
    if self.__samplers is None:
      self.reset()
    if not args:
      args = {0}, # Default to randomly sampling variables as scalars
    elif len(args) == 1 and type(args[0]) is int and 'stop' not in kwds:
      kwds.update({'stop': args[0]})
      args = {0},
    stop = None if 'stop' not in kwds else kwds.pop('stop')
    sampler = sample_generator(self, 
                               len(self.__samplers), 
                               *args, 
                               stop=stop, 
                               **kwds)
    self.__samplers.append(sampler)
    self.__counter[sampler] = 0
    self.__last[sampler] = None
    return sampler

#-------------------------------------------------------------------------------
  def walk(self, sampler, stop=None):
    gen_vars = inspect.getgeneratorlocals(sampler)
    assert 'stop' in gen_vars, \
        'Sampler must be a sampler_generator instance returned by SP.sampler()'
    if stop is None and gen_vars['stop'] is None:
      warnings.warn(
        "No stop specification set - this walk may proceed indefinitely")
    if stop is None: 
      return collections.deque([sample for sample in sampler])
    steps = collections.deque()
    for sample in sampler:
      if len(steps) >= stop:
        break
      steps.append(sample)
    return steps

#-------------------------------------------------------------------------------



""" A utility module for stochasptic process classes """

import numpy as np
from probayes.vtypes import isscalar
from probayes.pscales import rescale, div_prob

#-------------------------------------------------------------------------------
def sample_generator(sp, sampler_id, *args, stop=None, **kwds):
  if stop is None:
    while True:
      yield sp.next(sampler_id, *args, **kwds)
  else:
    while sp.get_counter(sampler_id) < stop:
      yield sp.next(sampler_id, *args, **kwds)
    else:
      sp.reset(sampler_id)

#-------------------------------------------------------------------------------
def metropolis_scores(opqr, pscale=None):
  pred, succ = opqr.o, opqr.p
  message = "No valid scalar probability distribution found"
  assert succ is not None, message
  assert isscalar(succ.prob), message 
  if pred is None:
    return None
  assert isscalar(pred.prob), "Preceding probability distribution non-scalar"
  return min(1., div_prob(succ.prob, pred.prob, pscale, pscale, pscale=1.))

#-------------------------------------------------------------------------------
def metropolis_thresh(*args, **kwds):
  return np.random.uniform(*args, **kwds)

#-------------------------------------------------------------------------------
def metropolis_update(stu):
  if stu.s is None or stu.s >= stu.t:
    return True
  return None

#-------------------------------------------------------------------------------
def hastings_scores(opqr, pscale=None):
  pred, succ, prop, revp = opqr.o, opqr.p, opqr.q, opqr.r
  message = "No valid scalar probability distribution found"
  assert succ is not None, message
  assert isscalar(succ.prob), message 
  if pred is None:
    return None
  assert isscalar(pred.prob), "Preceding probability non-scalar"
  if prop is None:
    return None
  else:
    assert isscalar(prop.prob), "Proposal probability non-scalar"
    prop = rescale(prop.prob, pscale, 1.)
    if prop <= 0.:
      return None
    if revp is None:
      return min(1., div_prob(succ.prob, pred.prob, pscale, pscale, pscale=1.))
    else:
      assert isscalar(revp.prob), "Reverse proposal probability non-scalar"
      revp = rescale(revp.prob, pscale, 1.)
      if revp <= 0.:
        return 1.
      return min(1., div_prob(succ.prob * prop, 
                              pred.prob * revp, 
                              pscale, pscale, pscale=1.))

#-------------------------------------------------------------------------------
def hastings_thresh(*args, **kwds):
  return metropolis_thresh(*args, **kwds)

#-------------------------------------------------------------------------------
def hastings_update(stu):
  return metropolis_update(stu)

#-------------------------------------------------------------------------------
def gibbs_scores(*args, **kwds):
  return np.nan

#-------------------------------------------------------------------------------
def gibbs_thresh(*args, **kwds):
  return np.nan

#-------------------------------------------------------------------------------
def gibbs_update(*args, **kwds):
  return True

#-------------------------------------------------------------------------------
MCMC_SAMPLERS = {
    'metropolis': (metropolis_scores, metropolis_thresh, metropolis_update),
    'hastings': (hastings_scores, hastings_thresh, hastings_update),
    'gibbs': (gibbs_scores, gibbs_thresh, gibbs_update),
                }

#-------------------------------------------------------------------------------



"""
Likelihood function utility module.
"""

#-------------------------------------------------------------------------------
import numpy as np
import collections
from probayes.rf_utils import slice_by_keyvals

#-------------------------------------------------------------------------------
def int_to_bin(num, min_dim=0):
  """Convert positive integer(s) to binary array(s).

  :param num: integer or 1D array of integers
  :param pad: optional minimum dimension (default 0):
  """

  if isinstance(num, (list, tuple)):
    num = np.array(num, dtype=int)
  if isinstance(num, np.ndarray):
    assert num.ndim == 1, "Input num must be integer or one dimensional array"
    min_dim = np.maximum(min_dim, len(np.binary_repr(np.max(num))))
    return np.vstack([int_to_bin(element, min_dim) for element in num])
  bin_str = np.binary_repr(num)
  if min_dim:
    bin_str = bin_str.zfill(min_dim)
  return np.array(np.array(list(bin_str)).astype(np.int8), bool)

#-------------------------------------------------------------------------------
def bin_to_int(arr, _multiple=None):
  """ Convert binary array(s) to positive integer(s)."""
  if isinstance(arr, (list, tuple)):
    arr = np.array(arr, dtype=int)
  assert isinstance(arr, np.ndarray) and arr.ndim and arr.ndim < 3, \
    "Input must be array type of not more than two dimensions"
  if arr.ndim == 2:
    if _multiple is None:
      _multiple = 1 << np.arange(arr.shape[1])[::-1]
    return np.hstack([bin_to_int(row, _multiple) for row in arr])
  if _multiple is None:
    _multiple = 1 << np.arange(arr.size)[::-1]
  return arr.dot(_multiple)

#-------------------------------------------------------------------------------
def bool_perm_freq(bool_2d, col_labels=None, base_freq=0):
  """ Returns a multidimensional array of counts of boolean permutations 
  contained in rows of bool_2d.

  :param bool_2d: a rXc 2D NumPy bool array with r examples of boolean
                  permutations of length c.
  :param col_labels: an optional list/tuple of keys for the labels.
  :param base_freq: baseline frequency to add to counts
                    (set to one for unit Laplacian smoothing).

  :return: counts or a tuple of (function, relative counts) if labels are given:

  counts: c-dimensional integer array of counts with each dimension ordered as
          [False, True] for each of the columns in bool_2d.
  function: returns relative frequency given keywords corresponding to labels

  """

  assert isinstance(bool_2d, np.ndarray) and bool_2d.ndim == 2 and \
      bool_2d.dtype == bool, "First input must be a 2D NumPy boolean array"
  
  rows, cols = bool_2d.shape
  counts = np.zeros([2] * cols, dtype=int)
  sequences = bool_2d.astype(int).tolist()
  for sequence in sequences:
    counts[tuple(sequence)] += 1
  if col_labels is None:
    return counts
  assert len(col_labels) == cols, \
      "Labels size {} incommensurate with input column number {}".format(
          len(col_labels), cols)
  dims = collections.OrderedDict()
  vals = collections.OrderedDict()
  ones = np.ones(cols, dtype=int)
  for dim, lbl in enumerate(col_labels):
    reshape = np.copy(ones)
    reshape[dim] = 2
    dims.update({lbl: dim})
    vals.update({lbl: np.array([False, True]).reshape(reshape)})
  if base_freq:
    base_freq += counts
  rel_freq = counts / rows

  def _func_bool_perm_freq(spec=None, **kwds):
    assert 'dims' in kwds, \
        "Output dimensionality not given - " + \
        "use SD.set_prob(function, passdims=True)"
    kwds = dict(kwds)
    spec_dims = kwds.pop('dims')
    if spec is None:
      spec = kwds
    else:
      assert not kwds, \
          "Unknown keywords: {}".format(kwds)
    return slice_by_keyvals(spec, vals, rel_freq, dims, spec_dims)
  
  return _func_bool_perm_freq, rel_freq

#-------------------------------------------------------------------------------



""" Provides Dependence class """

#-------------------------------------------------------------------------------
import collections
import networkx as nx
from probayes.variable import Variable
from probayes.field import Field
from probayes.cf import CF
from probayes.functional import Functional
NX_DIRECTED_GRAPH = nx.DiGraph

#-------------------------------------------------------------------------------
class Dependence (NX_DIRECTED_GRAPH, Field):
  """ A dependence is field with an explicit directed conditionality. The 
  dependences are represented as a graph, representing variables as vertices, 
  and edges for their corresponding inter-relations. 
  
  Direct conditional dependences across groups of variables are set using 
  conditional functions that inter-relate fields. This can be performed via 
  the implicit architectural interface or explicit dependency interface 
  (self.add_deps()).

  Modification of a dependence functional can be made from an instance, only
  affecting its correspondence leaf vertices. Intermediate dependence functions
  can be set only from dependence predecessors.

  For convenience to define deltas, Dependence supports defaulting of proposal 
  objects without actually supporting proposals.
  """

  # Protected
  _arch = None             # Implicit archectectural configuration
  _leafs = None            # Field of variables that do not condition others
  _roots = None            # Field of variables not dependent on others
  _stems = None            # OrderedDict of latent RVs
  _preds = None            # Field of variables that are predecessors to leafs
  _def_prop_obj = None     # Default value for prop_obj
  _prop_obj = None         # Object referencing propositional conditions
  _variable_cls = Variable # Variable class
  _field_cls = Field       # Field class
  _subfields = None        # Convenience dictionary for the roots and leafs RFs
  _is_stochastic = False   # Flag of stochasticity status

#------------------------------------------------------------------------------- 
  def __init__(self, *args):
    """ Initialises the Dependence with Variables, Fields, or Dependences. 
    See def_deps().; """
    NX_DIRECTED_GRAPH.__init__(self)
    self.def_deps(*args)

#-------------------------------------------------------------------------------
  @property
  def preds(self):
    return self._preds

  @property
  def deps(self):
    return self._deps

  @property
  def arch(self):
    return self._arch

  def def_deps(self, *args):
    """ Defaults the dependence with variables, fields, or dependence arguments.

    :param args: each arg may be a variable, field, or dependence with the 
                 dependence chain with running right to left. If one argument is
                 a dependence, then all arguments must comprise of dependence 
                 instances.
    """
    self._deps = None
    self._arch = None
    if not args:
      return
    arg_aredep = [isinstance(arg, type(self)) for arg in args]

    # Absence of Dependence instances means at most a single direct dependency
    if not any(arg_aredep):
      self._arch = self
      fields = [None] * len(args)
      for i, arg in enumerate(args):
        if isinstance(arg, self._variable_cls):
          fields[i] = self._field_cls(arg)
        elif isinstance(arg, self._field_cls):
          fields[i] = arg
        else:
          raise TypeError("Unrecognised input argument type: {}".format(
              type(arg)))

      # Declare vertices, adding edges if there are multiple arguments
      if len(fields) == 1:
        variables = list(fields[0].varlist)
        for var in variables:
          NX_DIRECTED_GRAPH.add_node(self, var)
        return self._refresh(fields[0])
      leafs_vars = fields[0].varlist
      roots_vars = fields[-1].varlist
      if len(fields) > 2:
        for field in fields[1:-1]:
          leafs_vars += list(field.varlist)

      # Add leafs then roots
      for var in leafs_vars:
        NX_DIRECTED_GRAPH.add_node(self, var)
      for var in roots_vars:
        NX_DIRECTED_GRAPH.add_node(self, var)

      # Add edges
      for roots_var in roots_vars:
        for leafs_var in leafs_vars:
          NX_DIRECTED_GRAPH.add_edge(self, roots_var, leafs_var)
      if len(fields) == 2:
        return self._refresh(fields[0], fields[1])
      return self._refresh(self._field_cls(*tuple(leafs_var)), fields[-1])

    # At this point, all arguments must be dependences
    assert all(arg_aredep), "Cannot mix Dependences with other input types"

    # Adding all vertices in forward order, starting with leafs then the rest
    for arg in args:
      leafs = list(arg.leafs.varlist)
      for leaf in leafs:
        NX_DIRECTED_GRAPH.add_node(self, leaf)
    for arg in args:
      variables = list(arg.nodes)
      for var in variables:
        NX_DIRECTED_GRAPH.add_node(self, var)

    # Adding all edges in reverse order
    for arg in args[::-1]:
      NX_DIRECTED_GRAPH.add_edges_from(self, arg.edges())

    # Explicit dependences to be added in reverse order with all implicit bets off
    deps = [arg.deps for arg in args[::-1]]
    if any(deps):
      self._arch = None
      [self.add_deps(dep) for dep in deps]
      return self._refresh()

    # Implicit dependences may either be in parallel or in series, but not both

    # Iterate args, add RV vertices, detect running roots/leafs and explicit
    run_leafs = [None] * len(args)
    run_roots = [None] * len(args)
    for i, arg in enumerate(args):
      run_leafs[i] = list(arg.leafs.varlist)
      run_roots[i] = list(arg.roots.varlist)

    # Detect for implicit serial dependences
    serial = len(args) > 1
    for i in range(len(args)-1):
      if run_roots[i] is None or set(run_roots[i]) != set(run_leafs[i+1]):
        serial = False
        break
    if serial:
      self._arch = list(args[::-1])
      leafs, roots = None, None
      for i, arg in enumerate(args):
        if i == 0:
          leafs = arg.leafs
          roots = arg.roots
        elif i == len(args) - 1:
          roots = arg.roots
      return self._refresh(leafs, roots)

    # Detect for implicit parallel dependences
    parallel = len(args) > 1 
    leafs = set()
    roots = []
    if parallel:
      for i, arg in enumerate(args):
        roots += run_roots[i]
        if not leafs:
          leafs = run_leafs[i]
        elif not len(run_roots[i]) or leafs != run_leafs[i]:
          parallel = False
          break
    if parallel:
      self._arch = tuple(args[::-1])
      leafs = args[0].leafs
      roots = self._field_cls(*tuple(roots))
      return self._refresh(leafs, roots)

    return self._refresh()

#-------------------------------------------------------------------------------
  @property
  def subfields(self):
    return self._subfields

  def _refresh(self, leafs=None, roots=None):
    """ Refreshes tree summaries, Dependence name and identity, and default 
    states. While roots and leafs are represented as Fields, stems are contained 
    within a single ordered dictionary to be flexible enough to accommodate 
    dependence  arborisations.

    :param leafs: sets default for leafs
    :param roots: sets default for roots
    """
    self._leafs = None
    self._stems = collections.OrderedDict()
    self._roots = None

    # If defaulting leafs, then assume a simple RF specification
    if leafs:
      assert type(leafs) is self._field_cls, \
          "Input leafs cannot be type {}".format(type(leafs))
      self._leafs = leafs
      self._roots = roots
      if self._roots:
        assert type(self._roots) is self._field_cls, \
          "Input roots cannot be type {}".format(type(roots))

    # Otherwise distinguish Variables belonging to leafs, roots, and stems
    else:
      leafs = [] # Field of vertices with no children/successors
      roots = [] # Field of vertices with no parents/predecessors (and not a leaf)
      self._stems = collections.OrderedDict()
      variables = list(self.nodes)
      for var in variables:
        parents = list(self.predecessors(var))
        children = list(self.successors(var))
        if parents and children:
          self._stems.update({var.name: var})
        elif children: # roots must have children
          roots += [var]
        else: # leafs can be parentless
          leafs += [var]
      self._leafs = self._field_cls(*tuple(leafs))
      if roots:
        self._roots = self._field_cls(*tuple(roots))

    # Evaluate leafs predecessors 
    self._preds = None
    if not self._stems:
      self._preds = None if not self._roots else self._roots
    else:
      preds = set()
      for leaf_var in self._leafs.varlist:
        [preds.add(pred_var) for pred_var in self.predecessors(leaf_var)]
      self._preds = self._field_cls(*tuple(preds))

    # Default functions for implicit architectures
    if self._arch: # TODO self._arch must always be defined by this point
      self._func = Functional(self._leafs, self._preds)

    # Set convenience subfields, and evaluate name and id from leafs and roots only
    super()._refresh()
    self._name = self._leafs.name
    self._id = self._leafs.id
    self._subfields = {'leafs': self._leafs}
    if self._roots:
      self._name += "|{}".format(self._roots.name)
      self._id += "_with_{}".format(self._roots.id)
      self._subfields.update({'roots': self._roots})
    self.eval_length()

    # Set the default proposal object and default the delta accordingly
    self._def_prop_obj = self._roots if self._roots is not None else self._leafs
    self._Delta = self._def_prop_obj.Delta
    self._delta_type = self._def_prop_obj._delta_type
    self.set_prop_obj(None) # this is for the instantiater to decide

#-------------------------------------------------------------------------------
  @property
  def func(self):
    return self._func

  def add_func(self, spec, func, *args, **kwds):
    assert self._func, \
        "Architectual specification not able to accommodate functionals"
    self._func.add_func(spec, func, *args, **kwds)

#-------------------------------------------------------------------------------
  def add_deps(self, out, inp=None, func=None, *args, **kwds):
    """ Adds a conditional dependence that conditions conditioning with respect
    to out being conditioned by inp by function func with *args and **kwds.
    """
    if self._arch:
      self.remove_edges_from(self.edges)
      self._arch = None
    if self._deps is None:
      self._deps = collections.OrderedDict()
    assert not self._prob, \
        "Cannot assign conditional dependencies alongside specified probability"
    if inp is None and func is None:
      for key, val in out.items():
        self._deps.update({key: val})
        self.add_edges_from(val)

    dep = CF(out, inp, func, *args, **kwds)
    dep_key = dep.ret_name()
    self._deps.update({dep_key: dep})
    out_keys = list(dep.ret_out().ret_keys(as_list=True))
    inp_keys = list(dep.ret_inp().ret_keys(as_list=True))
    for out_key in out_keys:
      for inp_key in inp_keys:
        self.add_edge(inp_key, out_key)
    return collections.OrderedDict({dep_key: self._deps[dep_key]})

#-------------------------------------------------------------------------------
  @property
  def prop_obj(self):
    return self._prop_obj

  def set_prop_obj(self, prop_obj=None):
    """ Sets the object used for assigning proposal distributions """
    self._prop_obj = prop_obj
    if self._prop_obj is None:
      return
    self._Delta = self._prop_obj.Delta
    self._delta_type = self._prop_obj._delta_type

#-------------------------------------------------------------------------------
  def leafs_roots(self, spec=None):
    """ Returns a proxy object from _subfields. """
    if spec is None:
      return self._subfields
    if not isinstance(spec, str) and spec not in self._subfields.values(): 
      return False
    if isinstance(spec, str):
      assert spec in self._subfields, \
          '{} absent from {}'.format(spec, self._name)
      return self._subfields[spec]
    return spec

#-------------------------------------------------------------------------------
  def set_delta(self, delta=None, *args, **kwds):
    _delta = self.leafs_roots(delta)
    if not _delta:
      return super().set_delta(delta, *args, **kwds)
    self.set_prop_obj(_delta)
    self._delta = _delta.delta
    self._delta_args = _delta._delta_args
    self._delta_kwds = _delta._delta_kwds
    self._delta_type = _delta._delta_type
    self._spherise = _delta._spherise
    return self._delta

#-------------------------------------------------------------------------------
  def eval_dist_name(self, values, suffix=None):
    if suffix is not None:
      return super().eval_dist_name(values, suffix)
    keys = self._keylist 
    vals = collections.OrderedDict()
    if not isinstance(vals, dict):
      vals.update({key: vals for key in keys})
    else:
      for key, val in values.items():
        if ',' in key:
          subkeys = key.split(',')
          for i, subkey in enumerate(subkeys):
            vals.update({subkey: val[i]})
        else:
          vals.update({key: val})
      for key in self._keylist:
        if key not in vals.keys():
          vals.update({key: None})
    marg_vals = collections.OrderedDict()
    if self._leafs:
      for key in self._leafs.keylist:
        if key in keys:
          marg_vals.update({key: vals[key]})
    cond_vals = collections.OrderedDict()
    if self._roots:
      for key in self._roots.keylist:
        if key in keys:
          cond_vals.update({key: vals[key]})
    marg_dist_name = self._leafs.eval_dist_name(marg_vals)
    cond_dist_name = '' if not self._roots else \
                     self._roots.eval_dist_name(cond_vals)
    dist_name = marg_dist_name
    if len(cond_dist_name):
      dist_name += "|{}".format(cond_dist_name)
    return dist_name

#-------------------------------------------------------------------------------
  def set_vars(self, *args):
    raise NotImplementedError()

#-------------------------------------------------------------------------------
  def add_var(self, rv):
    raise NotImplementedError()

#-------------------------------------------------------------------------------
  def eval_marg_prod(self, samples):
    raise NotImplementedError()

#-------------------------------------------------------------------------------
  def eval_deps(self, *args, _skip_parsing=False, **kwds):
    if self._deps is None:
      return None
    vals = self.parse_args(*args, **kwds) if not _skip_parsing else args[0]
    for key, val in self._deps.items():
      output = val(vals)
      evals, dims = None, None
      if isinstance(output, dict):
        evals = output
      elif not isinstance(output, tuple):
        raise TypeError("Unrecognised type {} for output for dependency {}".\
                        format(type(output), key))
      else: 
        evals = output[0] 
        assert isinstance(evals, dict),\
            "Unrecognised dependency evaluation output type: {}".format(
                type(evals))
        assert len(output) < 3, \
            "Maximum for 3 outputs from dependency evaluation"
        for argout in output:
          if isinstance(argout, dict):
            assert dims is None, "Output ambiguous for dimensionality"
            dims = argout
    return vals

#-------------------------------------------------------------------------------
  def eval_func(self, *args, _skip_parsing=False, **kwds):
    """ Evaluates leaf values from immediate predecessors """
    assert len(self._func), \
        "No functions defined for functional {}".format(self._func)
    vals = self.parse_args(*args, **kwds) if not _skip_parsing else args[0]

    # Can only evaluate architectures
    if not self._arch: 
      return vals

    # Iconics should be evaluable directly:
    if self._func.isiconic:
      pass

#-------------------------------------------------------------------------------
  def evaluate(self, *args, _skip_parsing=False, **kwds):
    """ Returns evaluation for Field() if there are no dependencies otherwise
    evaluation is based on functionals defined for each Dependences.
    """
    assert self._leafs, "No leaf stochastic random variables defined"
    if not self._arch or not self._func or not len(self._func):
      return super().evaluate(*args, _skip_parsing=_skip_parsing, **kwds)
    

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ Currently calls Field.__call__ """
    return super().__call__(*args, **kwds)

#-------------------------------------------------------------------------------
  def step(self, *args, **kwds):
    prop_obj = self._prop_obj
    if prop_obj is None:
      return super().step(*args, **kwds)
    prop_obj = prop_obj or self._def_prop_obj
    return prop_obj.step(*args, **kwds)

#-------------------------------------------------------------------------------



'''
A simple functional wrapper without the big guns of func_tools,
making use of the 'order' keyword to map arguments and keywords.
This makes 'order' a disallowed keyword. The 'delta' keyword is
similarly used to may differences in form var_name'-var_name,
making also a disallowed keyword.

Func may be a tuple of callable/uncallable functions
'''
import numpy as np
import scipy.stats
import collections
from probayes.vtypes import isscalar, isunitary

#-------------------------------------------------------------------------------
""" 
Scipy calls is supported a special case for following call convention:"
func() pdf
func[0]() pdf
func[1]() logpdf
func[2]() cdf
func[3]() logcdf
func[4]() rvs
"""
SCIPY_STATS_MVAR = {scipy.stats._multivariate.multi_rv_generic}

#------------------------------------------------------------------------------- 
def is_scipy_stats_mvar(arg, scipy_stats_mvar=SCIPY_STATS_MVAR):
  """ Returns a boolean for whether arg is an instance of scipy multivariate """
  return isinstance(arg, tuple(scipy_stats_mvar))

#-------------------------------------------------------------------------------
class Func:
  """ A functional wrapper to enable object representations as an uncallable
  array, a callable function, or a tuple of callable functions

  :example:
  >>> from probayes.func import Func
  >>> hw = Func("Hello world!")
  >>> print(hw())
  Hello World!
  >>> inc = Func(lambda x: x+1)
  >>> print(inc(2.)
  3.0
  >>> inc_dec = Func( (lambda x:x+1, lambda x:x-1) )
  >>> print(inc_dec[0](3.))
  4.0
  >>> print(inc_dec[1](3.))
  2.0
  """

  # Protected
  _func = None
  _args = None
  _kwds = None

  # Private
  __ismulti = None
  __isscalar = None # not of interest to Func but to objects that call Func
  __isscipy = None
  __callable = None
  __scipyobj = None
  __scipycalls = None
  __order = None
  __delta = None
  __index = None

#-------------------------------------------------------------------------------
  def __init__(self, func=None, *args, **kwds):
    """ Initialises Func according to object in func, which may be an uncallable
    object, a callable function, or a tuple of callable functions. See set_func()
    """
    self.set_func(func, *args, **kwds)
    
#-------------------------------------------------------------------------------
  def set_func(self, func=None, *args, **kwds):
    """ Set the Func instance's function object.

    :param func: an uncallable object, callable function, or tuple of functions
    :param *args: arguments to pass onto callables
    :param **kwds: keywords to pass onto callables

    Note that the following two reserved keywords are disallowed:

    'order': which instead denotes a dictionary of remappings.
    'delta': which instead denotes a mapping of differences.

    A special case is made for func instances that are Scipy multivariate objects.
    """

    self._func = func
    self._args = tuple(args)
    self._kwds = dict(kwds)
    self.__order = None
    self.__delta = None
    self.__callable = None
    self.__scipyobj = None
    self.__isscipy = False

    # Sanity check func
    if self._func is None:
      assert not args and not kwds, "No optional args without a function"
    self.__ismulti = isinstance(self._func, tuple)
    self.__isscalar = False
    if not self.__ismulti:
      self.__callable = callable(self._func)
      if not self.__callable:
        assert not args and not kwds, "No optional args with uncallable function"
        self.__isscalar = isscalar(self._func)
    else:
      func_callable = [callable(func) for func in self._func]
      func_isscalar = [isscalar(func) for func in self._func]
      assert len(set(func_callable)) < 2, \
          "Cannot mix callable and uncallable functions"
      assert len(set(func_isscalar)) < 2, \
          "Cannot mix scalars and nonscalars"
      if len(func_callable):
        self.__callable = func_callable[0]
        self.__isscalar = func_isscalar[0]
      if not self.__callable:
        assert not args and not kwds, "No optional args with uncallable function"
    if is_scipy_stats_mvar(self._func):
      self.__isscipy = True
      self.__scipyobj = self._func(*args, **kwds)
      self.__scipycalls = {
                           0: self.__scipyobj.pdf,
                           1: self.__scipyobj.logpdf,
                           2: self.__scipyobj.cdf,
                           3: self.__scipyobj.logcdf,
                           4: self.__scipyobj.rvs,
                          }
    if 'order' in self._kwds:
      self.set_order(self._kwds.pop('order'))
    if 'delta' in self._kwds:
      self.set_delta(self._kwds.pop('delta'))

#-------------------------------------------------------------------------------
  def set_order(self, order=None):
    """ Sets an order remapping dictionary for functional calls in which
    keyword arguments are mapped to position (in numeric) or rekeyed (if str).
    """
    self.__order = order
    if self.__order is None:
      return
    assert self.__delta is None, "Cannot set both order and delta"
    assert self.__scipyobj is None, \
        "Optional 'order' keyword prohibited for scipy objectts"
    self._check_mapping(self.__order)

#-------------------------------------------------------------------------------
  def set_delta(self, delta=None):
    """ Sets a difference remapping dictionary for functional calls in which
    keyword arguments are mapped to position (in numeric) or rekeyed (if str).
    """
    self.__delta = delta
    if self.__delta is None:
      return
    assert self.__order is None, "Cannot set both order and delta"
    assert self.__scipyobj is None, \
        "Optional 'delta' keyword prohibited for scipy objectts"
    self._check_mapping(self.__delta)

#-------------------------------------------------------------------------------
  def _check_mapping(self, mapping=None):
    """ Perform sanity checkings on mapping dictionary """
    if mapping is None:
      return
    # Used to sanity-check mapping dicts e.g. order and delta
    assert isinstance(mapping, dict), \
        "Mapping must be a dictionary type, not {}".format(type(mapping))
    key_list = list(mapping.keys())
    ind_list = list(mapping.values())
    keys = []
    inds = []
    for key, ind in zip(key_list, ind_list):
      keys.append(key)
      if type(ind) is int:
        inds.append(ind)
      elif ind is None:
        pass
      elif not isinstance(ind, str):
        raise TypeError("Cannot interpret index specification value: {}".ind)
    indset = set(inds)
    if len(indset):
      assert indset == set(range(min(indset), max(indset)+1)), \
          "Index specification non_sequitur: {}".format(indset)

#-------------------------------------------------------------------------------
  def ret_func(self):
    """ Returns function argument set by set_func() """
    return self._func

#-------------------------------------------------------------------------------
  def ret_args(self):
    """ Returns args as set by set_func() """
    return self._args

#-------------------------------------------------------------------------------
  def ret_kwds(self):
    """ Returns kwds as set by set_func() """
    return self._kwds

#-------------------------------------------------------------------------------
  def ret_callable(self):
    """ Returns boolean flag as to whether func is callable """
    return self.__callable

#-------------------------------------------------------------------------------
  def ret_isscalar(self):
    """ Returns boolean flag as to whether func is a scalar """
    return self.__isscalar

#-------------------------------------------------------------------------------
  def ret_ismulti(self):
    """ Returns boolean flag as to whether func comprises a multiple """
    return self.__ismulti

#-------------------------------------------------------------------------------
  def ret_isscipy(self):
    """ Returns boolean flag as to whether func comprises a scipy object """
    return self.__isscipy

#-------------------------------------------------------------------------------
  def ret_scipyobj(self):
    """ Returns scipy object if specified """
    return self.__scipyobj

#-------------------------------------------------------------------------------
  def ret_order(self):
    """ Returns order object if set """
    return self.__order

#-------------------------------------------------------------------------------
  def ret_delta(self):
    """ Returns delta object if set """
    return self.__delta

#-------------------------------------------------------------------------------
  def _call(self, *args, **kwds):
    """ Private call used by the wrapped Func interface.
    (see __call__ and __getitem__).
    """

    """
    # For debugging:
    argsin = args
    kwdsin = kwds
    """

    # Handle scipy objects separately
    if self.__isscipy:
      return self._call_scipy(*args, **kwds)

    # Check for indexing and reset if necessary
    func = self._func
    if self.__index is not None:
      func = func[self.__index]
      self.__index = None

    # Non-callables
    if not self.__callable:
      assert not args and not kwds, "No optional args with uncallable function"
      return func

    # Callables order-free
    if len(args) == 1 and isinstance(args[0], dict):
      args, kwds = (), {**kwds, **dict(args[0])}
    if self._args:
      args = tuple(list(self._args) + list(args))
    if self._kwds:
      kwds = {**kwds, **self._kwds}
    if not self.__order and not self.__delta:

      """
      # For debugging:
      argsout = args
      kwdsout = kwds
      import pdb; pdb.set_trace()
      """

      return func(*args, **kwds)

    # Append None to args according to mapping index specification
    n_args = len(args)
    mapping = self.__order or self.__delta
    for val in mapping.values():
      if type(val) is int:
        n_args = max(n_args, val+1)
    args = list(args)
    while len(args) < n_args:
      args.append(None)

    # Callables with order wrapper
    if self.__order:
      for key, val in self.__order.items():
        if type(val) is int:
          args[val] = kwds.pop(key)
        elif val is None:
          kwds.pop(key)
        elif isinstance(val, str):
          kwds.update({val: kwds.pop(key)})
        else:
          raise TypeError("Unrecognised order key: val type: {}:{}".\
                          format(key, val))
      return func(*tuple(args), **kwds)

    # Callables with delta wrapper
    for key, val in self.__delta.items():
      if key[-1] != "'":
        value = kwds.pop(key)
      else:
        value = kwds.pop(key) - kwds.pop(key[:-1])
      if type(val) is int:
        args[val] = value
      elif val is None:
        pass
      elif isinstance(val, str):
        kwds.update({val: value})
      else:
        raise TypeError("Unrecognised delta key: val type: {}:{}".\
                        format(key, val))
    return func(*tuple(args), **kwds)

#-------------------------------------------------------------------------------
  def _call_scipy(self, *args, **kwds):
    """ Private call used by the wrapped Func interface for scipy objects.
    (see __call__ and __getitem__).
    """
    index = 0
    if self.__index is not None:
      index = self.__index
      self.__index = None
    if index < 4:
      if len(args) == 1 and isinstance(args[0], dict):
        args = [np.ravel(val) for val in args[0].values()]
      elif not len(args) and len(kwds):
        args = list(collections.OrderedDict(**kwds).values())
        kwds = {}
      if isinstance(args, list) and not np.all([isunitary(arg) for arg in args]):
        args = args[::-1]
        if len(args) > 2:
          args = args[1:] + [args[0]]
        args = tuple(args)
      return self.__scipycalls[index](np.stack(np.meshgrid(*args), axis=-1), 
                                      **kwds)
    return self.__scipycalls[index](*args, **kwds)

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
   """ Wrapper call to the function with optional inclusion of additional
   args and kwds. """

   assert not self.__ismulti or self.__isscipy, \
       "Cannot call with multiple func, use Func[]"
   return self._call(*args, **kwds)

#-------------------------------------------------------------------------------
  def __getitem__(self, spec=None):
   r""" Calls the $i$th function from the Func tuple where the spec is $i$ """
   if spec is None:
     return self._func
   assert self.__ismulti or self.__isscipy, \
     "Cannot index without single func, use Func()"
   self.__index = spec
   return self._call

#-------------------------------------------------------------------------------
  def __len__(self):
    """ Returns the number of functions in the tuple set by set_func() """
    if not self.__ismulti:
      return None
    return len(self._func)

#-------------------------------------------------------------------------------



""" Provides stochastic dependence class SD. SD-related utility functions are
provided in sd_utils.py """
#-------------------------------------------------------------------------------
import numpy as np
import collections
from probayes.dependence import Dependence
from probayes.rv import RV
from probayes.rf import RF
from probayes.pd_utils import product
from probayes.sd_utils import desuffix, get_suffixed, arch_prob

#-------------------------------------------------------------------------------
class SD (Dependence, RF):
  """ A stochastic dependence is a random field that accommodates directed 
  conditionality according to one or more conditional probability distribution 
  functions. The dependences are represented as a graph, representing
  RVs as vertices and edges for their corresponding inter-relations. 
  
  Direct conditional dependences across groups of RVs are set using conditional 
  functions that inter-relate RFs. This can be performed via the implicit
  architectural interface (using SD.set_prob()) or explicit dependency
  interface (self.add_deps()).
  """

  # Protected
  _opqr = None         # (p(pred), p(succ), q(succ|pred), q(pred|succ))
  _tran_obj = None     # Object referencing transitional conditions
  _variable_cls = RV   # Variable class
  _field_cls = RF      # Field class
  _unit_prob = None    # Flag for single RV probability
  _unit_tran = None    # Flag for single RV transitional
  _is_stochastic = True  # Flag of stochasticity status

  # Private
  __sub_cfs = None     # Dictionary of conditional functions
  __sym_tran = None    # Flag to denote symmetrical conditionals

#------------------------------------------------------------------------------- 
  def __init__(self, *args):
    """ Initialises the SD with RVs, RFs, or SDs. See def_deps() """
    Dependence.__init__(self, *args)
    self.set_prob()

#-------------------------------------------------------------------------------
  @property
  def opqr(self):
    return self._opqr

  def _refresh(self, leafs=None, roots=None):
    """ Refreshes tree summaries, SD name and identity, and default states. 
    While roots and leafs are represented as RFs, stems are contained within a
    single ordered dictionary to be flexible enough to accommodate dependence 
    arborisations.

    :param leafs: sets default for leafs
    :param roots: sets default for roots
    """
    self.__sym_tran = False
    super()._refresh(leafs, roots)

    # Determine unit RVRF
    self._opqr = collections.namedtuple(self._id, ['o', 'p', 'q', 'r'])
    self._unit_prob = False
    self._unit_tran = False
    if self._nvars == 1:
      var = self._varlist[0]
      self._unit_prob = self._prob is None and var.prob is not None
      self._unit_tran = self._tran is None and var.tran is not None

#-------------------------------------------------------------------------------
  def set_prob(self, prob=None, *args, **kwds):
    """ Sets the joint probability with optional arguments and keywords.

    :param prob: may be a scalar, array, or callable function.
    :param pscale: represents the scale used to represent probabilities.
    :param *args: optional arguments to pass if prob is callable.
    :param **kwds: optional keywords to pass if prob is callable.
    """
    if prob is not None:
      assert self._deps is None, \
          "Cannot specify probabilities alongside deps conditional dependencies"
    prob = super().set_prob(prob, *args, **kwds)
    if prob is not None or not isinstance(self._arch, (list, tuple)):
      return prob
    return super().set_prob(arch_prob, arch=self._arch, passdims=True)

#-------------------------------------------------------------------------------
  def set_prop(self, prop=None, *args, **kwds):
    _prop = self.leafs_roots(prop)
    if not _prop:
      return super().set_prop(prop, *args, **kwds)
    self.set_prop_obj(_prop)
    self._prop = _prop._prop
    return self._prop

#-------------------------------------------------------------------------------
  def set_tran(self, tran=None, *args, **kwds):
    _tran = self.leafs_roots(tran)
    if not _tran:
      self._tran_obj = self
      return super().set_tran(tran, *args, **kwds)
    self._tran_obj = _tran
    self.set_prop_obj(self._tran_obj)
    self._tran = _tran.tran
    return self._tran

#-------------------------------------------------------------------------------
  def set_tfun(self, tfun=None, *args, **kwds):
    _tfun = self.leafs_roots(tfun)
    if not _tfun:
      self._tran_obj = self
      return super().set_tfun(tfun, *args, **kwds)
    self._tran_obj = _tfun
    self.set_prop_obj(_tfun)
    self._tfun = _tfun.tfun
    return self._tfun

#-------------------------------------------------------------------------------
  def eval_deps(self, *args, _skip_parsing=False, **kwds):
    if self._deps is None:
      return None
    vals = self.parse_args(*args, **kwds) if not _skip_parsing else args[0]
    for key, val in self._deps.items():
      output = val(vals)
      evals, dims, prob = None, None, None
      if isinstance(output, dict):
        evals = output
      elif not isinstance(output, tuple):
        raise TypeError("Unrecognised type {} for output for dependency {}".\
                        format(type(output), key))
      else: 
        evals = output[0] 
        assert isinstance(evals, dict),\
            "Unrecognised dependency evaluation output type: {}".format(
                type(evals))
        assert len(output) < 3, \
            "Maximum for 3 outputs from dependency evaluation"
        for argout in output:
          if isinstance(argout, dict):
            assert dims is None, "Output ambiguous for dimensionality"
            dims = argout
          elif isinstance(argout, np.ndarray):
            assert prob is None, "Output ambiguous for probabilities"
            prob = argout
    return vals

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
    """ Like RF.__call__ but optionally takes 'joint' keyword """
    if not self.anystoch:
      return Dependence.__call__(self, *args, **kwds)
    if not self._nvars:
      return None
    joint = False if 'joint' not in kwds else kwds.pop('joint')
    dist = super().__call__(*args, **kwds)
    if not joint:
      return dist
    vals = dist('cond')
    cond_dist = self._roots(vals)
    joint_dist = product(cond_dist, dist)
    return joint_dist

#-------------------------------------------------------------------------------
  def propose(self, *args, **kwds):
    prop_obj = self._prop_obj
    if prop_obj is None and (self._tran is not None or self._prop is not None):
      return super().propose(*args, **kwds)
    prop_obj = prop_obj or self._def_prop_obj
    return prop_obj.propose(*args, **kwds)

#-------------------------------------------------------------------------------
  def parse_pred_args(self, *args):
    if self._tran_obj == self:
      return self.parse_args(*args)
    if len(args) == 1 and isinstance(args[0], dict):
      arg = args[0]
      keyset = self._tran_obj.keyset
      pred = collections.OrderedDict({key: val for key, val in arg.items() 
                                               if key in keyset})
      return self._tran_obj.parse_args(pred)
    return self._tran_obj.parse_args(*args)

#-------------------------------------------------------------------------------
  def sample(self, *args, **kwds):
    """ A function for unconditional and conditional sampling. For conditional
    sampling, use RF.set_delta() to set the delta specification. if neither
    set_prob() nor set_tran() are set, then opqr inputs are disallowed and this
    function outputs a normal __call__(). Otherwise this function returns a 
    namedtuple-generated opqr object that can be accessed using opqr.p or 
    opqr[1] for the probability distribution and opqr.q or opqr[2] for the 
    proposal. Unavailable values are set to None. 
    
    If using set_prop() the output opqr comprises:

    opqr.o: None
    opqr.p: Probability distribution 
    opqr.q: Proposition distribution
    opqr.r: None

    If using set_tran() the output opqr comprises:

    opqr.o: Probability distribution for predecessor
    opqr.p: Probability distribution for successor
    opqr.q: Proposition distribution (successor | predecessor)
    opqr.r: None [for now, reserved for proposition (predecessor | successor)]

    If inputting and opqr object using set_prop(), the values for performing any
    delta operations are taken from the entered proposition distribution. If using
    set_prop(), optional keyword flag suffix=False may be used to remove prime
    notation in keys.

    An optional argument args[1] can included in order to input a dictionary
    of values beyond outside the proposition distribution required to evaluate
    the probability distribution.
    """
    if not args: # Default to randomly sampling variable scalars
      args = {0},
    assert len(args) < 3, "Maximum of two positional arguments"
    if self._tran is None and self._tfun is None and not self._unit_tran:
      if self._prop is None:
        assert not isinstance(args[0], self._opqr),\
            "Cannot input opqr object with neither set_prob() nor set_tran() set"
        return self.__call__(*args, **kwds)
      return self._sample_prop(*args, **kwds)
    return self._sample_tran(*args, **kwds)

#-------------------------------------------------------------------------------
  def _sample_prop(self, *args, **kwds):

    # Non-opqr argument requires no parsing
    if not isinstance(args[0], self._opqr):
      prop = self.propose(args[0], **kwds)

    # Otherwise parse:
    else:
      assert args[0].q is not None, \
          "An input opqr argument must contain a non-None value for opqr.q"
      vals = desuffix(args[0].q)
      prop = self.propose(vals, **kwds)

    # Evaluation of probability
    vals = desuffix(prop)
    if len(args) > 1:
      assert isinstance(args[1], dict),\
          "Second argument must be dictionary type, not {}".format(
              type(args[1]))
      vals.update(args[1])
    call = self.__call__(vals, **kwds)

    return self._opqr(None, call, prop, None)

#-------------------------------------------------------------------------------
  def _sample_tran(self, *args, **kwds):
    assert 'suffix' not in kwds, \
        "Disallowed keyword 'suffix' when using set_tran()"

    # Original probability distribution, proposal, and revp defaults to None
    orig = None
    prop = None
    revp = None

    # Non-opqr argument requires no parsing
    if not isinstance(args[0], self._opqr):
      prop = self.step(args[0], **kwds)

    # Otherwise parse successor:
    else:
      dist = args[0].q
      orig = args[0].p
      assert dist is not None, \
          "An input opqr argument must contain a non-None value for opqr.q"
      vals = get_suffixed(dist)
      prop = self.step(vals, **kwds)

    # Evaluate reverse proposal if transition function not symmetric
    if not self._sym_tran and not self._unit_tran:
      revp = self.reval_tran(prop)

    # Extract values evaluating probability
    vals = get_suffixed(prop)
    if len(args) > 1:
      assert isinstance(args[1], dict),\
          "Second argument must be dictionary type, not {}".format(
              type(args[1]))
      vals.update(args[1])
    prob = self.__call__(vals, **kwds)

    return self._opqr(orig, prob, prop, revp)

#-------------------------------------------------------------------------------



# Utility module for RF objects

#-------------------------------------------------------------------------------
import collections
import numpy as np
from probayes.vtypes import isscalar
from probayes.pscales import iscomplex, rescale, prod_rule, prod_pscale

#-------------------------------------------------------------------------------
def rv_prod_rule(*args, rvs, pscale=None):
  """ Returns the probability product treating all rvs as independent.
  Values (=args[0]) are keyed by RV name and rvs are a list of RVs.
  """
  values = args[0]
  pscales = [rv.pscale for rv in rvs]
  pscale = pscale or prod_pscale(pscales)
  use_logs = iscomplex(pscale)
  probs = [rv.eval_prob(values[rv.name]) for rv in rvs]
  prob, pscale = prod_rule(*tuple(probs),
                           pscales=pscales,
                           pscale=pscale)

  # This section below is there just to play nicely with conditionals
  if len(args) > 1:
    if use_logs:
      prob = rescale(prob, pscale, 0.j)
    else:
      prob = rescale(prob, pscale, 1.)
    for arg in args[1:]:
      if use_logs:
        offs, _ = rv_prod_rule(arg, rvs=rvs, pscale=0.j)
        prob = prob + offs
      else:
        coef, _ = rv_prod_rule(arg, rvs=rvs, pscale=1.)
        prob = prob * coef
    if use_logs:
      prob = prob / float(len(args))
      prob = rescale(prob, 0.j, pscale)
    else:
      prob = prob ** (1. / float(len(args)))
      prob = rescale(prob, 1., pscale)
  return prob, pscale

#-------------------------------------------------------------------------------
def call_scipy_prob(func, pscale, *args, **kwds):
  index = 1 if iscomplex(pscale) else 0
  return func[index](*args, **kwds)

#-------------------------------------------------------------------------------
def sample_cond_cov(*args, cond_cov=None, unknown=None, **kwds):
    kwds = dict(kwds)
    cond_pdf = False if 'cond_pdf' not in kwds else kwds.pop('cond_pdf')
    assert cond_cov, "coveig object mandatory"
    if len(args) == 1 and isinstance(args[0], dict):
      vals = dict(args[0])
      if unknown is not None:
        vals[unknown] = None
      args = [np.array(val) if val is not None else {0} \
              for val in vals.values()]
    elif not len(args) and len(kwds):
      vals = dict(kwds)
      if unknown is not None:
        kwds[unknown] = {0}
      args = list(kwds.values())
    return cond_cov.interp(*tuple(args), cond_pdf=cond_pdf)

#-------------------------------------------------------------------------------
def slice_by_keyvals(spec, vals, prob, vals_dims=None, spec_dims=None):
  """ Slices prob by values of spec in vals.

  :param spec: dictionary of {key:val} to match with vals.
  :param vals: dictionary of {key:val} describing prob.
  :param prob: a multidimensional NumPy array to slice.
  :param vals_dims: dictionary of dimensional decription for vals.
  :param spec_dims: dictionary of dimensional decription for spec.

  If vals_dims and/or spec_dims are not entered, they are 'guessed' from vals
  and spec respectively, but correct guessing is not assured. Non-none dimensions
  for vals_dims and spec_dims must be mutually ordered monotically by key.
  """

  # Check for consistent keys
  keys = list(spec.keys())
  assert set(keys) == set(vals.keys()), "Keys for spec and vals unmatched"

  # Function to default dimensions if not given
  def dims_from_vals(vals_dict):
    if not isinstance(vals_dict, dict):
      raise TypeError("Dictionary type expected")
    dims = collections.OrderedDict()
    run_dim = 0
    for key, val in vals_dict.items():
      if isscalar(val):
        dims.update({key: None})
      else:
        assert val.size == np.product(vals.shape), \
            "Multiple non-singleton dimensions: {}".format(val.size)
        if val.size > 1:
          run_dim = np.argmax(val.shape)
        dims.update({key: run_dim})
        run_dim += 1

  # Default spec_dims and vals_dims if not given 
  if spec_dims is None:
    spec_dims = dims_from_vals(spec)
  else:
    assert set(spec.keys()) == set(spec_dims.keys()), \
        "Keys for spec and spec_dims unmatched"
  if vals_dims is None:
    vals_dims = dims_from_vals(vals)
  else:
    assert set(vals.keys()) == set(vals_dims.keys()), \
        "Keys for spec and spec_dims unmatched"
  
  # Determine maximum dimensionality of input
  vals_ndim = 0
  for dim in vals_dims.values():
    if dim:
      vals_ndim = max(vals_ndim, dim)

  # Determine maximum dimensionality of output from spec and spec_dims
  spec_ndim = 0
  for key, dim in spec_dims.items():
    if dim:
      spec_ndim = max(spec_ndim, dim)
    if not isscalar(spec[key]):
      spec_ndim = max(spec[key].ndim, dim)

  # Check for monotonic ordering of dimensions
  dims = [dim for dim in vals_dims.values() if dim is not None]
  if len(dims) > 1:
    assert np.min(np.diff(dims)) > 0, "Dimensionality not monotically ordered"
  dims = [dim for dim in spec_dims.values() if dim is not None]
  if len(dims) > 1:
    assert np.min(np.diff(dims)) > 0, "Dimensionality not monotically ordered"

  # Evaluate reshape and slices from matches between spec and vals
  reshape = [1] * spec_ndim
  slices = [slice(None) for _ in range(vals_ndim+1)]
  for i, key in enumerate(keys):
    if vals_dims[key] is None: # If source is scalar
      if not isscalar(spec[key]):
        assert np.all(spec[key] == vals[key]), \
          "Cannot slice by multiple values"
      elif spec[key] != vals[key]:
        return np.empty(np.zeros(len(dims), dtype=int), dtype=float)
      else:
        pass
    if spec_dims[key] is None: # If target is scalar
      dim = vals_dims[key]
      match = np.ravel(vals[key]) == spec[key]
      n_matches = match.sum()
      if n_matches == 0:
        slices[dim] = slice(0, 0)
      elif n_matches == 1:
        slices[dim] = np.nonzero(match)[0]
      else:
        raise ValueError("Non-unique matches found")
    else:
      assert np.all(np.ravel(vals[key]) == np.ravel(spec[key])), \
          "Ambiguous specification with values mismatch"
      dim = spec_dims[key]
      reshape[dim] = vals[key].size
  return prob[tuple(slices)].reshape(reshape)

#-------------------------------------------------------------------------------



# flake8: noqa
__version__ = '0.0.5'
from probayes.constants import NEARLY_POSITIVE_ZERO, \
                               NEARLY_POSITIVE_INF, \
                               NEARLY_NEGATIVE_INF, \
                               LOG_NEARLY_POSITIVE_INF, \
                               COMPLEX_ZERO
from probayes.vtypes import OO
from probayes.named_dict import NamedDict
from probayes.icon import Icon
from probayes.expr import Expr
from probayes.variable import Variable, variables
from probayes.prob import Prob
from probayes.rv import RV, RVs
from probayes.field import Field
from probayes.rf import RF
from probayes.dependence import Dependence
from probayes.sd import SD
from probayes.sp import SP
from probayes.cf import CF
from probayes.distribution import Distribution
from probayes.pd import PD
from probayes.pd_utils import product, summate, iterdict, read_dist, write_dist
from probayes.pd_utils import serialise, deserialise, read_serialised, write_serialised
from probayes.likelihoods import bool_perm_freq
from probayes.expression import Expression
from probayes.functional import Functional
from probayes.sympy_prob import SympyProb
